from supriya.intervals.Moment import Moment
from supriya.system import SupriyaObject


cdef class _CInterval:

    cdef readonly float start_offset
    cdef readonly float stop_offset
    cdef readonly object original_interval

    def __cinit__(self, start_offset, stop_offset, original_interval):
        self.start_offset = start_offset
        self.stop_offset = stop_offset
        self.original_interval = original_interval

    def __richcmp__(_CInterval self, _CInterval other, int op):
        if op == 0:  # <
            if self.start_offset < other.start_offset:
                return True
            if self.start_offset > other.start_offset:
                return False
            if self.stop_offset < other.stop_offset:
                return True
            return False
        if op == 2:  # ==
            if self.start_offset != other.start_offset:
                return False
            if self.stop_offset != other.stop_offset:
                return False
            if self.original_interval != other.original_interval:
                return False
            return True

    @classmethod
    def from_interval(cls, interval):
        start_offset = float(interval.start_offset)
        stop_offset = float(interval.stop_offset)
        return cls(
            start_offset,
            stop_offset,
            interval,
            )

    cpdef bint intersects_interval(_CInterval self, _CInterval other):
        return (
            (
                other.start_offset <= self.start_offset and
                self.start_offset < other.stop_offset
            ) or (
                self.start_offset <= other.start_offset and
                other.start_offset < self.stop_offset
                )
            )


cdef class _CNode:

    cdef public int balance
    cdef public int height
    cdef public _CNode left_child
    cdef public int node_start_index
    cdef public int node_stop_index
    cdef public object payload
    cdef public _CNode right_child
    cdef public float start_offset
    cdef public float stop_offset_high
    cdef public float stop_offset_low
    cdef public int subtree_start_index
    cdef public int subtree_stop_index

    def __init__(self, start_offset):
        self.balance = 0
        self.height = 0
        self.left_child = None
        self.node_start_index = -1
        self.node_stop_index = -1
        self.payload = []
        self.right_child = None
        self.start_offset = start_offset
        self.stop_offset_high = start_offset
        self.stop_offset_low = start_offset
        self.subtree_start_index = -1
        self.subtree_stop_index = -1


cdef class IntervalTreeDriverEx:
    """
    A mutable always-sorted collection of intervals.
    """

    ### CLASS VARIABLES ###

    cdef public object _root_node

    ### INITIALIZER ###

    def __init__(self, intervals=None):
        self._root_node = None
        self.update(intervals or [])

    ### SPECIAL METHODS ###

    def __contains__(self, interval):
        if not self._is_interval(interval):
            raise ValueError(interval)
        candidates = self.find_intervals_starting_at(interval.start_offset)
        result = interval in candidates
        return result

    def __getitem__(self, i):
        if isinstance(i, int):
            if self._root_node is None:
                raise IndexError
            if i < 0:
                i = self._root_node.subtree_stop_index + i
            if i < 0 or self._root_node.subtree_stop_index <= i:
                raise IndexError
            cinterval = self._recurse_getitem_by_index(self._root_node, i)
            return cinterval.original_interval
        elif isinstance(i, slice):
            if self._root_node is None:
                return []
            indices = i.indices(self._root_node.subtree_stop_index)
            start, stop = indices[0], indices[1]
            cintervals = self._recurse_getitem_by_slice(self._root_node, start, stop)
            return self._unbox_cintervals(cintervals)
        raise TypeError('Indices must be integers or slices, got {}'.format(i))

    def __iter__(self):
        cdef _CNode current
        cdef _CInterval cinterval
        stack = []
        current = self._root_node
        while True:
            while current is not None:
                stack.append(current)
                current = current.left_child
            if not stack:
                return
            current = stack.pop()
            for i in range(len(current.payload)):
                cinterval = current.payload[i]
                yield cinterval.original_interval
            while current.right_child is None and stack:
                current = stack.pop()
                for i in range(len(current.payload)):
                    cinterval = current.payload[i]
                    yield cinterval.original_interval
            current = current.right_child

    def __len__(self):
        if self._root_node is None:
            return 0
        return self._root_node.subtree_stop_index

    ### PRIVATE METHODS ###

    cdef _CNode _insert_node(
        self,
        _CNode node,
        float start_offset,
        ):
        cdef _CNode child_node
        if node is None:
            return _CNode(start_offset)
        if start_offset < node.start_offset:
            child_node = self._insert_node(node.left_child, start_offset)
            self._set_node_left_child(node, child_node)
        elif node.start_offset < start_offset:
            child_node = self._insert_node(node.right_child, start_offset)
            self._set_node_right_child(node, child_node)
        return self._rebalance(node)

    cdef void _insert_interval(
        self,
        _CInterval cinterval,
        ):
        cdef _CNode node
        self._root_node = self._insert_node(
            self._root_node,
            cinterval.start_offset,
            )
        node = self._search(self._root_node, cinterval.start_offset)
        node.payload.append(cinterval)
        node.payload.sort(key=lambda x: x.stop_offset)

    @staticmethod
    def _is_interval(expr):
        if hasattr(expr, 'start_offset') and hasattr(expr, 'stop_offset'):
            return True
        return False

    cdef _CNode _rebalance(
        self,
        _CNode node,
        ):
        if node is None:
            return None
        if 1 < node.balance:
            if 0 <= node.right_child.balance:
                node = self._rotate_right_right(node)
            else:
                node = self._rotate_right_left(node)
        elif node.balance < -1:
            if node.left_child.balance <= 0:
                node = self._rotate_left_left(node)
            else:
                node = self._rotate_left_right(node)
        return node

    cdef object _recurse_find_intervals_intersecting_offset(
        self,
        _CNode node,
        float offset,
        ):
        cdef _CInterval cinterval
        result = []
        if node is None:
            return result
        if node.start_offset <= offset < node.stop_offset_high:
            subresult = self._recurse_find_intervals_intersecting_offset(
                node.left_child, offset)
            result.extend(subresult)
            for i in range(len(node.payload)):
                cinterval = node.payload[i]
                if offset < cinterval.stop_offset:
                    result.append(cinterval)
            subresult = self._recurse_find_intervals_intersecting_offset(
                node.right_child, offset)
            result.extend(subresult)
        elif offset <= node.start_offset:
            subresult = self._recurse_find_intervals_intersecting_offset(
                node.left_child, offset)
            result.extend(subresult)
        return result

    cdef object _recurse_find_intervals_intersecting_interval(
        self,
        _CNode node,
        _CInterval cinterval,
        ):
        cdef _CInterval candidate_interval, node_interval
        result = []
        if node is None:
            return result
        node_interval = _CInterval(
            node.start_offset,
            node.stop_offset_high,
            node,
            )
        if cinterval.intersects_interval(node_interval):
            subresult = self._recurse_find_intervals_intersecting_interval(
                node.left_child, cinterval)
            result.extend(subresult)
            for i in range(len(node.payload)):
                candidate_interval = node.payload[i]
                if candidate_interval.intersects_interval(cinterval):
                    result.append(candidate_interval)
            subresult = self._recurse_find_intervals_intersecting_interval(
                node.right_child, cinterval)
            result.extend(subresult)
        elif (cinterval.start_offset <= node.start_offset) or \
            (cinterval.stop_offset <= node.start_offset):
            subresult = self._recurse_find_intervals_intersecting_interval(
                node.left_child, cinterval)
            result.extend(subresult)
        return result

    cdef object _recurse_find_intervals_stopping_at(
        self,
        _CNode node,
        float offset,
        ):
        cdef _CInterval cinterval
        result = []
        if node is None:
            return result
        if node.stop_offset_low <= offset <= node.stop_offset_high:
            for i in range(len(node.payload)):
                cinterval = node.payload[i]
                if cinterval.stop_offset == offset:
                    result.append(cinterval)
            if node.left_child is not None:
                result.extend(self._recurse_find_intervals_stopping_at(
                    node.left_child, offset))
            if node.right_child is not None:
                result.extend(self._recurse_find_intervals_stopping_at(
                    node.right_child, offset))
        return result

    cdef float _recurse_get_offset_after(
        self,
        _CNode node,
        float offset,
        int depth=1,
    ):
        if node is None:
            # print(("    " * depth) + "None")
            return offset
        result = offset
        # print(("    " * depth) + "Node: {} {} {}".format(node.start_offset, node.stop_offset_low, node.stop_offset_high))
        if node.start_offset > offset:
            # print(("    " * depth) + "After")
            result = node.start_offset
            if offset < node.stop_offset_low < result:
                result = node.stop_offset_low
            # print(("    " * depth) + "Checking Left")
            candidate = self._recurse_get_offset_after(node.left_child, offset, depth + 1)
            if offset != result:
                if offset < candidate < result:
                    result = candidate
            elif offset < candidate:
                result = candidate
        elif node.start_offset <= offset and (
            node.stop_offset_high > offset or node.stop_offset_high > offset
        ):
            # print(("    " * depth) + "Before or Equal")
            for interval in node.payload:
                if interval.stop_offset > offset:
                    result = interval.stop_offset
                    # print(("    " * depth) + "Found: {}".format(result))
                    break
            # print(("    " * depth) + "Checking Right")
            candidate = self._recurse_get_offset_after(node.right_child, offset, depth + 1)
            if offset != result:
                if offset < candidate < result:
                    result = candidate
            elif offset < candidate:
                result = candidate
            # print(("    " * depth) + "Checking Left")
            candidate = self._recurse_get_offset_after(node.left_child, offset, depth + 1)
            if offset != result:
                if offset < candidate < result:
                    result = candidate
            elif offset < candidate:
                result = candidate
        else:
            # print(("    " * depth) + "Nope")
            pass
        # print(("    " * depth) + "Result: {}".format(result))
        return result

    cdef _CNode _recurse_get_start_offset_after(
        self,
        _CNode node,
        float offset,
    ):
        result = None
        if node is None:
            return result
        if node.start_offset <= offset and node.right_child:
            result = self._recurse_get_start_offset_after(
                node.right_child, offset)
        elif offset < node.start_offset:
            result = self._recurse_get_start_offset_after(
                node.left_child, offset) or node
        return result

    cdef _CNode _recurse_get_start_offset_before(
        self,
        _CNode node,
        float offset,
    ):
        result = None
        if node is None:
            return result
        if node.start_offset < offset:
            result = self._recurse_get_start_offset_before(
                node.right_child, offset) or node
        elif offset <= node.start_offset and node.left_child:
            result = self._recurse_get_start_offset_before(
                node.left_child, offset)
        return result

    cdef _CInterval _recurse_getitem_by_index(
        self,
        _CNode node, 
        int index,
        ):
        if node.node_start_index <= index < node.node_stop_index:
            return node.payload[index - node.node_start_index]
        elif node.left_child and index < node.node_start_index:
            return self._recurse_getitem_by_index(node.left_child, index)
        elif node.right_child and node.node_stop_index <= index:
            return self._recurse_getitem_by_index(node.right_child, index)

    cdef object _recurse_getitem_by_slice(
        self,
        _CNode node,
        int start,
        int stop,
        ):
        cdef int node_start, node_stop
        result = []
        if node is None:
            return result
        if start < node.node_start_index and node.left_child:
            result.extend(self._recurse_getitem_by_slice(
                node.left_child, start, stop))
        if start < node.node_stop_index and node.node_start_index < stop:
            node_start = start - node.node_start_index
            if node_start < 0:
                node_start = 0
            node_stop = stop - node.node_start_index
            result.extend(node.payload[node_start:node_stop])
        if node.node_stop_index <= stop and node.right_child:
            result.extend(self._recurse_getitem_by_slice(
                node.right_child, start, stop))
        return result

    cdef _CNode _remove_node(
        self,
        _CNode node,
        float start_offset,
        ):
        cdef _CNode next_node, child_node
        if node is None:
            return
        if node.start_offset == start_offset:
            if node.left_child and node.right_child:
                next_node = node.right_child
                while next_node.left_child is not None:
                    next_node = next_node.left_child
                node.start_offset = next_node.start_offset
                node.payload = next_node.payload
                self._set_node_right_child(node, self._remove_node(
                    node.right_child,
                    next_node.start_offset,
                    ))
            else:
                node = node.left_child or node.right_child
        elif start_offset < node.start_offset:
            child_node = self._remove_node(node.left_child, start_offset)
            self._set_node_left_child(node, child_node)
        elif node.start_offset < start_offset:
            child_node = self._remove_node(node.right_child, start_offset)
            self._set_node_right_child(node, child_node)
        return self._rebalance(node)

    cdef void _remove_interval(
        self,
        _CInterval cinterval,
        ):
        cdef _CNode node
        node = self._search(self._root_node, cinterval.start_offset)
        if node is None:
            return
        if cinterval in node.payload:
            node.payload.remove(cinterval)
        if not node.payload:
            self._root_node = self._remove_node(
                self._root_node,
                cinterval.start_offset,
                )

    cdef _CNode _rotate_left_left(
        self,
        _CNode node
        ):
        next_node = node.left_child
        self._set_node_left_child(node, next_node.right_child)
        self._set_node_right_child(next_node, node)
        return next_node

    cdef _CNode _rotate_left_right(
        self,
        _CNode node,
        ):
        self._set_node_left_child(
            node, self._rotate_right_right(node.left_child))
        next_node = self._rotate_left_left(node)
        return next_node

    cdef _CNode _rotate_right_left(
        self,
        _CNode node,
        ):
        self._set_node_right_child(
            node, self._rotate_left_left(node.right_child))
        next_node = self._rotate_right_right(node)
        return next_node

    cdef _CNode _rotate_right_right(
        self,
        _CNode node,
        ):
        next_node = node.right_child
        self._set_node_right_child(node, next_node.left_child)
        self._set_node_left_child(next_node, node)
        return next_node

    cdef _CNode _search(
        self,
        _CNode node,
        float start_offset,
        ):
        if node is None:
            return None
        if node.start_offset == start_offset:
            return node
        elif (
            node.left_child is not None and 
            start_offset < node.start_offset
            ):
            return self._search(node.left_child, start_offset)
        elif (
            node.right_child is not None and 
            node.start_offset < start_offset
            ):
            return self._search(node.right_child, start_offset)
        return None

    cdef void _set_node_left_child(
        self,
        _CNode node,
        _CNode left_child,
        ):
        node.left_child = left_child
        self._update_node_height(node)

    cdef void _set_node_right_child(
        self,
        _CNode node,
        _CNode right_child,
        ):
        node.right_child = right_child
        self._update_node_height(node)

    cdef object _unbox_cintervals(
        self,
        cintervals,
        ):
        cdef _CInterval cinterval
        cintervals.sort(key=lambda x: (x.start_offset, x.stop_offset))
        return [cinterval.original_interval for cinterval in cintervals]

    cdef void _update_node_height(
        self,
        _CNode node,
        ):
        cdef int left_height, right_height
        left_height = -1
        right_height = -1
        if node.left_child is not None:
            left_height = node.left_child.height
        if node.right_child is not None:
            right_height = node.right_child.height
        node.height = max(left_height, right_height) + 1
        node.balance = right_height - left_height

    cdef void _update_indices(
        self,
        _CNode node,
        int parent_stop_index,
        ):
        if node is None:
            return
        if node.left_child is not None:
            self._update_indices(
                node.left_child,
                parent_stop_index,
                )
            node.node_start_index = node.left_child.subtree_stop_index
            node.subtree_start_index = node.left_child.subtree_start_index
        elif parent_stop_index == -1:
            node.node_start_index = 0
            node.subtree_start_index = 0
        else:
            node.node_start_index = parent_stop_index
            node.subtree_start_index = parent_stop_index
        node.node_stop_index = node.node_start_index + len(node.payload)
        node.subtree_stop_index = node.node_stop_index
        if node.right_child is not None:
            self._update_indices(
                node.right_child,
                node.node_stop_index,
                )
            node.subtree_stop_index = node.right_child.subtree_stop_index

    cdef _CNode _update_offsets(
        self, 
        _CNode node,
        ):
        cdef _CNode child_node
        cdef _CInterval cinterval
        cdef float stop_offset_low, stop_offset_hight
        if node is None:
            return
        stop_offset_low = float('inf')
        stop_offset_high = float('-inf')
        for i in range(len(node.payload)):
            cinterval = node.payload[i]
            if cinterval.stop_offset < stop_offset_low:
                stop_offset_low = cinterval.stop_offset
            if cinterval.stop_offset > stop_offset_high:
                stop_offset_high = cinterval.stop_offset
        if node.left_child is not None:
            child_node = self._update_offsets(
                node.left_child,
                )
            if child_node.stop_offset_low < stop_offset_low:
                stop_offset_low = child_node.stop_offset_low
            if stop_offset_high < child_node.stop_offset_high:
                stop_offset_high = child_node.stop_offset_high
        if node.right_child is not None:
            child_node = self._update_offsets(
                node.right_child,
                )
            if child_node.stop_offset_low < stop_offset_low:
                stop_offset_low = child_node.stop_offset_low
            if stop_offset_high < child_node.stop_offset_high:
                stop_offset_high = child_node.stop_offset_high
        node.stop_offset_low = stop_offset_low
        node.stop_offset_high = stop_offset_high
        return node

    ### PUBLIC METHODS ###

    def add(self, interval):
        if not self._is_interval(interval):
            # raise ValueError(interval)
            return
        cinterval = _CInterval.from_interval(interval)
        self._insert_interval(cinterval)
        self._update_indices(self._root_node, -1)
        self._update_offsets(self._root_node)

    def find_intervals_starting_at(self, offset):
        results = []
        node = self._search(self._root_node, offset)
        if node is None:
            return results
        return self._unbox_cintervals(node.payload)

    def find_intervals_stopping_at(self, offset):
        cintervals = self._recurse_find_intervals_stopping_at(
            self._root_node, offset)
        return self._unbox_cintervals(cintervals)

    def find_intervals_intersecting_offset(self, offset):
        offset = float(offset)
        cintervals = self._recurse_find_intervals_intersecting_offset(
            self._root_node, offset)
        return self._unbox_cintervals(cintervals)

    def find_intervals_intersecting_interval(self, interval):
        cinterval = _CInterval.from_interval(interval)
        assert isinstance(cinterval, _CInterval)
        cintervals = self._recurse_find_intervals_intersecting_interval(
            self._root_node, cinterval)
        return self._unbox_cintervals(cintervals)

    def get_offset_after(self, offset):
        # print("Searching for: {}".format(offset))
        result = None
        if self._root_node is not None:
            result = self._recurse_get_offset_after(self._root_node, offset)
            if result == offset:
                result = None
        return result

    def get_start_offset_after(self, offset):
        node = self._recurse_get_start_offset_after(self._root_node, offset)
        if node is None:
            return None
        return node.start_offset

    def get_start_offset_before(self, offset):
        node = self._recurse_get_start_offset_before(self._root_node, offset)
        if node is None:
            return None
        return node.start_offset

    def index(self, interval):
        assert self._is_interval(interval)
        cinterval = _CInterval.from_interval(interval)
        node = self._search(self._root_node, cinterval.start_offset)
        if node is None:
            raise ValueError('{} not in interval tree.'.format(interval))
        if cinterval not in node.payload:
            raise ValueError('{} not in interval tree.'.format(interval))
        index = node.payload.index(cinterval) + node.node_start_index
        return index

    def remove(self, interval):
        if interval not in self:
            raise ValueError(interval)
        cinterval = _CInterval.from_interval(interval)
        self._remove_interval(cinterval)
        self._update_indices(self._root_node, -1)
        self._update_offsets(self._root_node)

    def update(self, intervals):
        # for interval in intervals:
        #     if not self._is_interval(interval):
        #         raise ValueError(interval)
        for interval in intervals:
            if not self._is_interval(interval):
                # raise ValueError(interval)
                continue
            cinterval = _CInterval.from_interval(interval)
            self._insert_interval(cinterval)
        self._update_indices(self._root_node, -1)
        self._update_offsets(self._root_node)



#!/usr/bin/env python
import pathlib
import platform

from setuptools import Extension, setup

package_name = "supriya"


def read_version():
    root_path = pathlib.Path(__file__).parent
    version_path = root_path / package_name / "_version.py"
    with version_path.open() as file_pointer:
        file_contents = file_pointer.read()
    local_dict = {}
    exec(file_contents, None, local_dict)
    return local_dict["__version__"]


if __name__ == "__main__":
    setup(
        author="Josiah Wolf Oberholtzer",
        author_email="josiah.oberholtzer@gmail.com",
        classifiers=[
            "Environment :: Console",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: MacOS",
            "Operating System :: POSIX",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Artistic Software",
            "Topic :: Multimedia :: Sound/Audio :: Sound Synthesis",
        ],
        description="A Python API for SuperCollider",
        extras_require={
            "docs": [
                "librosa",
                "matplotlib >= 3.3.0",
                "sphinx-design",
                "sphinx-immaterial",
                "sphinx-immaterial @ git+https://git@github.com/josiah-wolf-oberholtzer/sphinx-immaterial@main#egg=sphinx-immaterial",
                "sphinxext-opengraph",
            ],
            "ipython": [
                "jupyter",
                "jupyter_contrib_nbextensions",
                "jupyter_nbextensions_configurator",
                "rise",
            ],
            "test": [
                "black",
                "flake8 >= 3.9.0",
                "isort >= 5.8.0",
                "librosa",
                "matplotlib >= 3.3.0",
                "mypy >= 0.900",
                "pytest >= 6.2.0",
                "pytest-asyncio >= 0.14.0",
                "pytest-cov >= 2.12.0",
                "pytest-helpers-namespace >= 2019.1.8",
                "pytest-mock >= 3.5.0",
                "pytest-rerunfailures >= 9.1.0",
                "pytest-timeout >= 1.4.0",
                "types-PyYAML",
                "types-docutils",
            ],
        },
        ext_modules=[
            Extension(
                "supriya.realtime.shm",
                include_dirs=[
                    "vendor",
                    "vendor/TLSF-2.4.6/src",
                    "vendor/supercollider/common",
                ],
                language="c++",
                libraries=["rt"] if platform.system() == "Linux" else [],
                optional=True,
                sources=["supriya/realtime/shm.pyx"],
            )
        ],
        include_package_data=True,
        install_requires=[
            "Cython >= 0.29.0",
            "PyYAML >= 5.4.0",
            "platformdirs >= 2.4.0",
            "tqdm >= 4.59.0",
            "uqbar >= 0.5.7",
            'typing-extensions; python_version<"3.8"',
        ],
        keywords=[
            "audio",
            "dsp",
            "music composition",
            "scsynth",
            "supercollider",
            "synthesis",
        ],
        license="MIT",
        long_description=pathlib.Path("README.md").read_text(),
        long_description_content_type="text/markdown",
        name=package_name,
        packages=[package_name],
        url=f"https://github.com/josiah-wolf-oberholtzer/{package_name}",
        version=read_version(),
        zip_safe=False,
    )



from supriya.system import SupriyaObject


class Moment(SupriyaObject):
    """
    A moment of intervals in a interval tree.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_interval_tree",
        "_overlap_intervals",
        "_start_intervals",
        "_start_offset",
        "_stop_intervals",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        interval_tree=None,
        overlap_intervals=None,
        start_intervals=None,
        start_offset=None,
        stop_intervals=None,
    ):
        self._interval_tree = interval_tree
        self._start_offset = start_offset
        self._start_intervals = start_intervals
        self._stop_intervals = stop_intervals
        self._overlap_intervals = overlap_intervals

    ### SPECIAL METHODS ###

    def __repr__(self):
        """
        Gets the repr of this moment.
        """
        return "<{}({} <<{}>>)>".format(
            type(self).__name__,
            str(self.start_offset),
            len(self.start_intervals) + len(self.overlap_intervals),
        )

    ### PUBLIC PROPERTIES ###

    @property
    def next_moment(self):
        """
        Gets the next moment in this moment's interval
        collection.
        """
        # TODO: This doesn't take into account stop offsets
        tree = self._interval_tree
        if tree is None:
            return None
        start_offset = tree.get_start_offset_after(self.start_offset)
        if start_offset is None:
            return None
        return tree.get_moment_at(start_offset)

    @property
    def next_start_offset(self):
        """
        Gets the next moment start offset in this moment's
        interval tree.
        """
        tree = self._interval_tree
        if tree is None:
            return None
        start_offset = tree.get_start_offset_after(self.start_offset)
        return start_offset

    @property
    def overlap_intervals(self):
        """
        Gets the intervals in this moment which overlap this
        moment's start offset.
        """
        return self._overlap_intervals

    @property
    def previous_moment(self):
        """
        Gets the previous moment in this moment's interval
        collection.
        """
        # TODO: This doesn't take into account stop offsets
        tree = self._interval_tree
        if tree is None:
            return None
        start_offset = tree.get_start_offset_before(self.start_offset)
        if start_offset is None:
            return None
        return tree.get_moment_at(start_offset)

    @property
    def previous_start_offset(self):
        """
        Gets the previous moment start offset in this moment's
        interval tree.
        """
        tree = self._interval_tree
        if tree is None:
            return None
        start_offset = tree.get_start_offset_before(self.start_offset)
        return start_offset

    @property
    def start_offset(self):
        """
        Gets this moment's start offset.
        """
        return self._start_offset

    @property
    def start_intervals(self):
        """
        Gets the intervals in this moment which start at this
        moment's start offset.
        """
        return self._start_intervals

    @property
    def stop_intervals(self):
        """
        Gets the intervals in this moment which stop at this
        moment's start offset.
        """
        return self._stop_intervals

    @property
    def interval_tree(self):
        """
        Gets this moment's interval tree.
        """
        return self._interval_tree



import hashlib
import pathlib
import shutil
import struct
import subprocess
from os import PathLike
from typing import Any, Optional, Tuple

import tqdm  # type: ignore
import uqbar.containers
import uqbar.io
import yaml
from uqbar.objects import new

import supriya
import supriya.realtime
import supriya.soundfiles
import supriya.system
from supriya import HeaderFormat, SampleFormat, scsynth
from supriya.exceptions import NonrealtimeOutputMissing, NonrealtimeRenderError
from supriya.system import SupriyaObject
from supriya.typing import HeaderFormatLike, SampleFormatLike


class SessionRenderer(SupriyaObject):
    """
    Renders non-realtime sessions as audio files.
    """

    ### INITIALIZER ###

    def __init__(
        self,
        session: "supriya.nonrealtime.sessions.Session",
        header_format: HeaderFormatLike = HeaderFormat.AIFF,
        print_transcript: bool = False,
        render_directory_path: Optional[PathLike] = None,
        sample_format: SampleFormatLike = SampleFormat.INT24,
        sample_rate: int = 44100,
        transcript_prefix: Optional[str] = None,
    ):
        self._session = session

        self._header_format = HeaderFormat.from_expr(header_format)

        if print_transcript:
            print_transcript = bool(print_transcript)
        self._print_transcript = print_transcript

        self._render_directory_path = (
            pathlib.Path(render_directory_path or supriya.output_path)
            .expanduser()
            .absolute()
        )

        self._sample_format = SampleFormat.from_expr(sample_format)

        self._sample_rate = int(sample_rate)

        transcript_prefix = transcript_prefix or None
        if transcript_prefix:
            transcript_prefix = str(transcript_prefix)
        self._transcript_prefix = transcript_prefix

        self._reset()

    ### PRIVATE METHODS ###

    def _build_datagram(self, osc_bundles):
        datagrams = []
        for osc_bundle in osc_bundles:
            datagram = osc_bundle.to_datagram(realtime=False)
            size = len(datagram)
            size = struct.pack(">i", size)
            datagrams.append(size)
            datagrams.append(datagram)
        datagram = b"".join(datagrams)
        return datagram

    def _build_file_path(self, datagram, input_file_path, session):
        # print('BUILDING FILE PATH')
        md5 = hashlib.md5()
        md5.update(datagram)
        hash_values = []
        if input_file_path is not None:
            hash_values.append(input_file_path)
        for value in (
            session.options.input_bus_channel_count,
            session.options.output_bus_channel_count,
            self.sample_rate,
            self.header_format,
            self.sample_format,
        ):
            hash_values.append(value)
        for value in hash_values:
            if not isinstance(value, str):
                value = str(value)
            value = value.encode()
            md5.update(value)
        md5 = md5.hexdigest()
        file_path = "session-{}.osc".format(md5)
        return pathlib.Path(file_path)

    def _build_render_command(
        self,
        input_file_path,
        output_file_path,
        session_osc_file_path,
        *,
        scsynth_path=None,
        server_options=None,
    ):
        cwd = pathlib.Path.cwd()
        scsynth_path = scsynth.find(scsynth_path)
        server_options = server_options or scsynth.Options()
        if session_osc_file_path.is_absolute():
            session_osc_file_path = session_osc_file_path.relative_to(cwd)
        parts = [scsynth_path, "-N", session_osc_file_path]
        if input_file_path:
            parts.append(input_file_path)
        else:
            parts.append("_")
        if output_file_path.is_absolute() and cwd in output_file_path.parents:
            output_file_path = output_file_path.relative_to(cwd)
        parts.append(output_file_path)
        parts.append(self.sample_rate)
        parts.append(self.header_format.name.lower())  # Must be lowercase.
        parts.append(self.sample_format.name.lower())  # Must be lowercase.
        server_options = server_options.as_options_string(realtime=False)
        if server_options:
            parts.append(server_options)
        command = " ".join(str(_) for _ in parts)
        return command

    def _build_render_yml(self, session_prefixes):
        session_prefixes = session_prefixes[:]
        render_data = {"render": session_prefixes.pop(), "source": None}
        if session_prefixes:
            render_data["source"] = list(reversed(session_prefixes))
        render_yaml = yaml.dump(render_data, default_flow_style=False, indent=4)
        return render_yaml

    def _build_xrefd_bundles(self, osc_bundles):
        extension = ".{}".format(self.header_format.name.lower())
        for osc_bundle in osc_bundles:
            for osc_message in osc_bundle.contents:
                contents = list(osc_message.contents)
                for i, x in enumerate(contents):
                    x = self._sessionable_to_session(x)
                    try:
                        if x not in self.renderable_prefixes:
                            continue
                    except TypeError:
                        continue
                    renderable_file_path = self.renderable_prefixes[x].with_suffix(
                        extension
                    )
                    contents[i] = str(renderable_file_path)
                osc_message.contents = tuple(contents)
        return osc_bundles

    def _build_dependency_graph_and_nonxrefd_osc_bundles_conditionally(
        self, expr, parent
    ):
        import supriya.nonrealtime

        expr = self._sessionable_to_session(expr)
        if isinstance(expr, supriya.nonrealtime.Session):
            if expr not in self.dependency_graph:
                self._build_dependency_graph_and_nonxrefd_osc_bundles(expr)
            self.dependency_graph.add(expr, parent=parent)
        elif hasattr(expr, "__render__"):
            self.dependency_graph.add(expr, parent=parent)

    def _build_dependency_graph_and_nonxrefd_osc_bundles(self, session, duration=None):
        input_ = session.input_
        if isinstance(input_, str):
            input_ = pathlib.Path(input_)
        input_ = self._sessionable_to_session(input_)
        non_xrefd_bundles = session._to_non_xrefd_osc_bundles(duration)
        self.compiled_sessions[session] = input_, non_xrefd_bundles
        if session is self.session:
            self.dependency_graph.add(session)
        self._build_dependency_graph_and_nonxrefd_osc_bundles_conditionally(
            input_, session
        )
        for non_xrefd_bundle in non_xrefd_bundles:
            for request in non_xrefd_bundle.contents:
                for x in request.contents:
                    self._build_dependency_graph_and_nonxrefd_osc_bundles_conditionally(
                        x, session
                    )

    def _call_subprocess(self, command):
        return subprocess.call(command, shell=True)

    def _stream_subprocess(self, command, session_duration):
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        previous_value = 0
        progress_bar = tqdm.tqdm(
            bar_format=(), total=int(session_duration * 1000), unit="ms"
        )
        with progress_bar:
            while True:
                output = process.stdout.readline()
                if not output:
                    if process.poll() is not None:
                        break
                    continue
                output = output.decode().strip()
                if output.startswith("nextOSCPacket"):
                    current_value = int(float(output.split()[-1]) * 1000)
                    difference = current_value - previous_value
                    progress_bar.update(difference)
                    previous_value = current_value
                elif output.startswith("FAILURE"):
                    if output.startswith("FAILURE IN SERVER /n_free Node"):
                        continue
                    progress_bar.write(output)
                elif output.startswith("start time 0"):
                    continue
                else:
                    progress_bar.write("WARNING: {}".format(output))
                    if output.startswith("alloc failed"):
                        return -6
        return process.poll()

    def _collect_prerender_tuples(self, session, duration=None):
        import supriya.nonrealtime

        self._build_dependency_graph_and_nonxrefd_osc_bundles(
            session, duration=duration
        )
        assert self.dependency_graph.is_acyclic()
        extension = ".{}".format(self.header_format.name.lower())
        for renderable in self.dependency_graph:
            if isinstance(renderable, supriya.nonrealtime.Session):
                result = self._collect_session_prerender_tuple(renderable, extension)
                prerender_tuple, renderable_prefix = result
            else:
                result = self._collect_renderable_prerender_tuple(renderable)
                prerender_tuple, renderable_prefix = result
            self.prerender_tuples.append(prerender_tuple)
            self.renderable_prefixes[renderable] = renderable_prefix
        return self.prerender_tuples

    def _collect_renderable_prerender_tuple(self, renderable):
        renderable_prefix = renderable._build_file_path().with_suffix("")
        return (renderable,), renderable_prefix

    def _collect_session_prerender_tuple(self, session, extension):
        input_, non_xrefd_bundles = self.compiled_sessions[session]
        osc_bundles = self._build_xrefd_bundles(non_xrefd_bundles)
        input_file_path = input_
        if input_ and input_ in self.renderable_prefixes:
            input_file_path = self.renderable_prefixes[input_]
            input_file_path = input_file_path.with_suffix(extension)
        if input_file_path:
            input_file_path = self.get_path_relative_to_render_path(
                input_file_path, self.render_directory_path
            )
            self.session_input_paths[session] = input_file_path
        datagram = self._build_datagram(osc_bundles)
        renderable_prefix = self._build_file_path(
            datagram, input_file_path, session
        ).with_suffix("")
        return (session, datagram, input_, osc_bundles), renderable_prefix

    def _render_datagram(
        self,
        session,
        input_file_path,
        output_file_path,
        session_osc_file_path,
        scsynth_path=None,
        **kwargs,
    ):
        relative_session_osc_file_path = session_osc_file_path
        if relative_session_osc_file_path.is_absolute():
            relative_session_osc_file_path = session_osc_file_path.relative_to(
                pathlib.Path.cwd()
            )
        self._report("Rendering {}.".format(relative_session_osc_file_path))
        if output_file_path.exists():
            self._report(
                "    Skipped {}. Output already exists.".format(
                    relative_session_osc_file_path
                )
            )
            return 0
        server_options = session._options
        server_options = new(server_options, **kwargs)
        memory_size = server_options.memory_size
        for factor in range(1, 6):
            command = self._build_render_command(
                input_file_path,
                output_file_path,
                session_osc_file_path,
                scsynth_path=scsynth_path,
                server_options=server_options,
            )
            self._report("    Command: {}".format(command))
            try:
                exit_code = self._stream_subprocess(command, session.duration)
            except KeyboardInterrupt:
                if output_file_path.exists():
                    output_file_path.unlink()
                raise
            server_options = new(
                server_options, memory_size=memory_size * (2**factor)
            )
            if exit_code == -6:
                self._report(
                    "    Out of memory. Increasing to {}.".format(
                        server_options.memory_size
                    )
                )
            else:
                self._report(
                    "    Rendered {} with exit code {}.".format(
                        relative_session_osc_file_path, exit_code
                    )
                )
                break
        return exit_code

    def _read(self, file_path, mode=""):
        try:
            with open(str(file_path), "r" + mode) as file_pointer:
                return file_pointer.read()
        except FileNotFoundError:
            return None

    def _report(self, message):
        if self.transcript_prefix:
            message = "{}{}".format(self.transcript_prefix, message)
        if self.print_transcript:
            print(message)
        self.transcript.append(message)

    def _reset(self):
        self._compiled_sessions = {}
        self._prerender_tuples = []
        self._session._transcript = self._transcript = []
        self._renderable_prefixes = {}
        self._dependency_graph = uqbar.containers.DependencyGraph()
        self._session_input_paths = {}
        self._sessionables_to_sessions = {}

    def _sessionable_to_session(self, expr):
        if hasattr(expr, "__session__"):
            if expr not in self._sessionables_to_sessions:
                self._sessionables_to_sessions[expr] = expr.__session__()
            return self._sessionables_to_sessions[expr]
        return expr

    def _write_datagram(self, file_path, new_contents):
        self._write(file_path, new_contents, mode="b")

    def _write_render_yml(self, file_path, render_yaml):
        self._write(file_path, render_yaml)

    def _write(self, file_path, new_contents, mode=""):
        cwd = pathlib.Path.cwd()
        relative_file_path = file_path
        if file_path.is_absolute() and cwd in file_path.parents:
            relative_file_path = file_path.relative_to(cwd)
        self._report("Writing {}.".format(relative_file_path))
        old_contents = self._read(file_path, mode=mode)
        if old_contents == new_contents:
            self._report(
                "    Skipped {}. File already exists.".format(relative_file_path)
            )
        else:
            with open(str(file_path), "w" + mode) as file_pointer:
                file_pointer.write(new_contents)
            self._report("    Wrote {}.".format(relative_file_path))

    ### PUBLIC METHODS ###

    def to_lists(self, duration=None):
        osc_bundles = self.to_osc_bundles(duration=duration)
        return [osc_bundle.to_list() for osc_bundle in osc_bundles]

    def to_osc_bundles(self, duration=None):
        self._collect_prerender_tuples(self.session, duration=duration)
        (session, datagram, input_file_path, osc_bundles) = self.prerender_tuples[-1]
        return osc_bundles

    @classmethod
    def get_path_relative_to_render_path(cls, target_path, render_path):
        target_path = pathlib.Path(target_path)
        render_path = pathlib.Path(render_path)
        try:
            return target_path.relative_to(render_path)
        except ValueError:
            pass
        target_path = pathlib.Path(target_path)
        render_path = pathlib.Path(render_path)
        target_path_parents = set(target_path.parents)
        render_path_parents = set(render_path.parents)
        common_parents = target_path_parents.intersection(render_path_parents)
        if not common_parents:
            return target_path
        common_parent = sorted(common_parents)[-1]
        target_path = target_path.relative_to(common_parent)
        render_path = render_path.relative_to(common_parent)
        parts = [".." for _ in render_path.parts] + [target_path]
        return pathlib.Path().joinpath(*parts)

    def render(
        self,
        output_file_path: Optional[PathLike] = None,
        debug: bool = False,
        duration: Optional[float] = None,
        build_render_yml: bool = False,
        scsynth_path: Optional[str] = None,
        **kwargs,
    ) -> Tuple[int, Any, pathlib.Path]:
        import supriya.nonrealtime

        if output_file_path is not None:
            output_file_path = pathlib.Path(output_file_path).expanduser().absolute()

        self._collect_prerender_tuples(self.session, duration=duration)
        assert self.prerender_tuples, self.prerender_tuples

        extension = ".{}".format(self.header_format.name.lower())
        visited_renderable_prefixes = []

        with uqbar.io.DirectoryChange(directory=str(self.render_directory_path)):
            for prerender_tuple in self.prerender_tuples:
                renderable = prerender_tuple[0]
                renderable_prefix = self.renderable_prefixes[renderable]
                visited_renderable_prefixes.append(
                    renderable_prefix.with_suffix("").name
                )
                relative_output_file_path = renderable_prefix.with_suffix(extension)
                if isinstance(renderable, supriya.nonrealtime.Session):
                    (session, datagram, input_, _) = prerender_tuple
                    osc_file_path = renderable_prefix.with_suffix(".osc")
                    input_file_path = self.session_input_paths.get(session)
                    self._write_datagram(osc_file_path, datagram)
                    exit_code = self._render_datagram(
                        session,
                        input_file_path,
                        relative_output_file_path,
                        osc_file_path,
                        scsynth_path=scsynth_path,
                        **kwargs,
                    )
                    if exit_code:
                        self._report("    SuperCollider errored!")
                        raise NonrealtimeRenderError(exit_code)
                else:
                    renderable.__render__(
                        output_file_path=relative_output_file_path,
                        print_transcript=self.print_transcript,
                    )

        final_rendered_file_path = (
            self.render_directory_path / relative_output_file_path
        )
        if not final_rendered_file_path.exists():
            self._report("    Output file is missing!")
            raise NonrealtimeOutputMissing(final_rendered_file_path)

        if output_file_path is not None:
            shutil.copy(final_rendered_file_path, output_file_path)

        if build_render_yml:
            output_directory = (output_file_path or final_rendered_file_path).parent
            render_yaml = self._build_render_yml(visited_renderable_prefixes)
            self._write_render_yml(output_directory / "render.yml", render_yaml)

        return (
            exit_code,
            self.transcript,
            output_file_path or final_rendered_file_path,
        )

    ### PUBLIC PROPERTIES ###

    @property
    def compiled_sessions(self):
        return self._compiled_sessions

    @property
    def dependency_graph(self):
        return self._dependency_graph

    @property
    def header_format(self) -> HeaderFormat:
        return self._header_format

    @property
    def prerender_tuples(self):
        return self._prerender_tuples

    @property
    def print_transcript(self) -> bool:
        return self._print_transcript

    @property
    def render_directory_path(self) -> pathlib.Path:
        return self._render_directory_path

    @property
    def renderable_prefixes(self):
        return self._renderable_prefixes

    @property
    def sample_format(self) -> SampleFormat:
        return self._sample_format

    @property
    def sample_rate(self) -> int:
        return self._sample_rate

    @property
    def session(self) -> "supriya.nonrealtime.sessions.Session":
        return self._session

    @property
    def session_input_paths(self):
        return self._session_input_paths

    @property
    def transcript(self):
        return self._transcript

    @property
    def transcript_prefix(self):
        return self._transcript_prefix



import bisect
from uuid import uuid4

from uqbar.objects import get_vars, new

from supriya.assets import synthdefs
from supriya.enums import CalculationRate

from .events import (
    BusAllocateEvent,
    BusFreeEvent,
    CompositeEvent,
    GroupAllocateEvent,
    NodeFreeEvent,
    NullEvent,
    SynthAllocateEvent,
)
from .patterns import Pattern


class BusPattern(Pattern):

    ### INITIALIZER ###

    def __init__(
        self, pattern, calculation_rate="audio", channel_count=1, release_time=0.25
    ):
        self._pattern = pattern
        self._calculation_rate = CalculationRate.from_expr(calculation_rate)
        self._channel_count = channel_count
        self._release_time = release_time

    ### PRIVATE METHODS ###

    def _adjust(self, expr, state):
        args, _, kwargs = get_vars(expr)
        updates = {}
        if hasattr(expr, "target_node") and expr.target_node is None:
            updates["target_node"] = state["group"]
        if hasattr(expr, "synthdef"):
            synthdef = getattr(expr, "synthdef") or synthdefs.default
            parameter_names = synthdef.parameter_names
            for name in ("in_", "out"):
                if name in parameter_names and kwargs.get(name) is None:
                    updates[name] = state["bus"]
        if updates:
            return new(expr, **updates)
        return expr

    def _iterate(self, state=None):
        return iter(self._pattern)

    def _setup_peripherals(self, state):
        rate = self._calculation_rate.name.lower()
        link_synthdef_name = f"system_link_{rate}_{self._channel_count}"
        starts = [
            BusAllocateEvent(
                calculation_rate=self._calculation_rate,
                channel_count=self._channel_count,
                id_=state["bus"],
            ),
            GroupAllocateEvent(id_=state["group"]),
            SynthAllocateEvent(
                add_action="ADD_AFTER",
                amplitude=1.0,
                fade_time=self._release_time,
                in_=state["bus"],
                synthdef=getattr(synthdefs, link_synthdef_name),
                target_node=state["group"],
                id_=state["link"],
            ),
        ]
        stops = [
            NodeFreeEvent(id_=state["link"]),
            NodeFreeEvent(id_=state["group"]),
            BusFreeEvent(id_=state["bus"]),
        ]
        if self._release_time:
            stops.insert(1, NullEvent(delta=self._release_time))
        return CompositeEvent(starts), CompositeEvent(stops)

    def _setup_state(self):
        return {"bus": uuid4(), "link": uuid4(), "group": uuid4()}

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self):
        return self._pattern.is_infinite


class FxPattern(Pattern):

    ### INITIALIZER ###

    def __init__(self, pattern, synthdef, release_time=0.25, **kwargs):
        self._pattern = pattern
        self._release_time = release_time
        self._synthdef = synthdef
        self._kwargs = kwargs

    ### PRIVATE METHODS ###

    def _iterate(self, state=None):
        return iter(self._pattern)

    def _setup_peripherals(self, state):
        starts = [
            SynthAllocateEvent(
                add_action="ADD_TO_TAIL",
                synthdef=self._synthdef,
                id_=state["synth"],
                **self._kwargs,
            )
        ]
        stops = [NodeFreeEvent(id_=state["synth"])]
        if self._release_time:
            stops.insert(0, NullEvent(delta=self._release_time))
        return CompositeEvent(starts), CompositeEvent(stops)

    def _setup_state(self):
        return {"synth": uuid4()}

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self):
        return self._pattern.is_infinite


class GroupPattern(Pattern):

    ### INITIALIZER ###

    def __init__(self, pattern, release_time=0.25):
        self._pattern = pattern
        self._release_time = release_time

    ### PRIVATE METHODS ###

    def _adjust(self, expr, state):
        updates = {}
        if hasattr(expr, "target_node") and expr.target_node is None:
            updates["target_node"] = state["group"]
        if updates:
            return new(expr, **updates)
        return expr

    def _iterate(self, state=None):
        return iter(self._pattern)

    def _setup_peripherals(self, state):
        starts = [GroupAllocateEvent(add_action="ADD_TO_HEAD", id_=state["group"])]
        stops = [NodeFreeEvent(id_=state["group"])]
        if self._release_time:
            stops.insert(0, NullEvent(delta=self._release_time))
        return CompositeEvent(starts), CompositeEvent(stops)

    def _setup_state(self):
        return {"group": uuid4()}

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self):
        return self._pattern.is_infinite


class ParallelPattern(Pattern):

    ### INITIALIZER ###

    def __init__(self, patterns):
        self._patterns = tuple(patterns)

    ### PRIVATE METHODS ###

    def _iterate(self, state=None):
        should_stop = False
        iterators = []
        for index, pattern in enumerate(self._patterns):
            iterators.append((0.0, index, iter(pattern)))
        while iterators:
            grouping_offset = iterators[0][0]
            events = []
            while iterators and iterators[0][0] == grouping_offset:
                offset, index, iterator = iterators.pop(0)
                try:
                    if should_stop:
                        event = iterator.send(should_stop)
                    else:
                        event = next(iterator)
                    events.append(event)
                    triple = (offset + event.delta, index, iterator)
                    insert_index = bisect.bisect_left(iterators, triple)
                    iterators.insert(insert_index, triple)
                except StopIteration:
                    pass
            if events:
                if iterators:
                    delta = iterators[0][0] - grouping_offset
                else:
                    delta = max(event.delta for event in events)
                if len(events) == 1:
                    sent = yield new(events[0], delta=delta)
                elif len(events) > 1:
                    sent = yield CompositeEvent(
                        [new(x, delta=0.0) for x in events], delta=delta
                    )
                if sent:
                    should_stop = True

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self):
        return any(pattern.is_infinite for pattern in self._patterns)



import collections

from supriya import CalculationRate
from supriya.synthdefs import MultiOutUGen, UGen


class DiskIn(MultiOutUGen):
    """
    Streams in audio from a file.

    ::

        >>> buffer_id = 23
        >>> disk_in = supriya.ugens.DiskIn.ar(
        ...     buffer_id=buffer_id,
        ...     channel_count=2,
        ...     loop=0,
        ... )
        >>> disk_in
        UGenArray({2})

    """

    _default_channel_count = 1
    _has_done_flag = True
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict([("buffer_id", None), ("loop", 0.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class DiskOut(UGen):
    """
    Records to a soundfile to disk.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.SinOsc.ar(frequency=[440, 442])
        >>> disk_out = supriya.ugens.DiskOut.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        >>> disk_out
        DiskOut.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("source", None)]
    )
    _unexpanded_input_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class VDiskIn(MultiOutUGen):
    """
    Streams in audio from a file, with variable rate.

    ::

        >>> buffer_id = 23
        >>> vdisk_in = supriya.ugens.VDiskIn.ar(
        ...     buffer_id=buffer_id,
        ...     channel_count=2,
        ...     loop=0,
        ...     rate=1,
        ...     send_id=0,
        ... )
        >>> vdisk_in
        UGenArray({2})

    """

    _default_channel_count = 1
    _has_done_flag = True
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("rate", 1), ("loop", 0), ("send_id", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)



import collections

from supriya import CalculationRate
from supriya.synthdefs import PureUGen, UGen


class COsc(PureUGen):
    """
    A chorusing wavetable oscillator.

    ::

        >>> cosc = supriya.ugens.COsc.ar(
        ...     beats=0.5,
        ...     buffer_id=23,
        ...     frequency=440,
        ... )
        >>> cosc
        COsc.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("frequency", 440.0), ("beats", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class DegreeToKey(PureUGen):
    """
    A signal-to-modal-pitch converter.`

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> degree_to_key = supriya.ugens.DegreeToKey.ar(
        ...     buffer_id=23,
        ...     octave=12,
        ...     source=source,
        ... )
        >>> degree_to_key
        DegreeToKey.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("source", None), ("octave", 12)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Impulse(PureUGen):
    """
    A non-band-limited single-sample impulse generator unit generator.

    ::

        >>> supriya.ugens.Impulse.ar()
        Impulse.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Index(PureUGen):
    """
    A clipping buffer indexer.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> index = supriya.ugens.Index.ar(
        ...     buffer_id=23,
        ...     source=source,
        ... )
        >>> index
        Index.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("source", None)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFCub(PureUGen):
    """
    A sine-like oscillator unit generator.

    ::

        >>> supriya.ugens.LFCub.ar()
        LFCub.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("initial_phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFGauss(PureUGen):
    """
    A non-band-limited gaussian function oscillator.

    ::

        >>> supriya.ugens.LFGauss.ar()
        LFGauss.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [
            ("duration", 1),
            ("width", 0.1),
            ("initial_phase", 0),
            ("loop", 1),
            ("done_action", 0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### INITIALIZER ###

    def __init__(
        self,
        calculation_rate=None,
        done_action=0,
        duration=1,
        initial_phase=0,
        loop=1,
        width=0.1,
    ):
        import supriya.synthdefs

        done_action = supriya.DoneAction.from_expr(done_action)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            done_action=done_action,
            duration=duration,
            initial_phase=initial_phase,
            loop=loop,
            width=width,
        )


class LFPar(PureUGen):
    """
    A parabolic oscillator unit generator.

    ::

        >>> supriya.ugens.LFPar.ar()
        LFPar.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("initial_phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFPulse(PureUGen):
    """
    A non-band-limited pulse oscillator.

    ::

        >>> supriya.ugens.LFPulse.ar()
        LFPulse.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("initial_phase", 0), ("width", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFSaw(PureUGen):
    """
    A non-band-limited sawtooth oscillator unit generator.

    ::

        >>> supriya.ugens.LFSaw.ar()
        LFSaw.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("initial_phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFTri(PureUGen):
    """
    A non-band-limited triangle oscillator unit generator.

    ::

        >>> supriya.ugens.LFTri.ar()
        LFTri.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("initial_phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Osc(PureUGen):
    """
    An interpolating wavetable oscillator.
    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", 0), ("frequency", 440.0), ("initial_phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class OscN(PureUGen):
    """
    A non-interpolating wavetable oscillator.
    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", 0), ("frequency", 440.0), ("initial_phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Select(PureUGen):
    """
    A signal selector.

    ::

        >>> sources = supriya.ugens.In.ar(bus=0, channel_count=8)
        >>> selector = supriya.ugens.Phasor.kr() * 8
        >>> select = supriya.ugens.Select.ar(
        ...     sources=sources,
        ...     selector=selector,
        ... )
        >>> select
        Select.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("selector", None), ("sources", None)]
    )
    _unexpanded_input_names = ("sources",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class SinOsc(PureUGen):
    """
    A sinusoid oscillator unit generator.

    ::

        >>> supriya.ugens.SinOsc.ar()
        SinOsc.ar()

    ::

        >>> print(_)
        synthdef:
            name: ...
            ugens:
            -   SinOsc.ar:
                    frequency: 440.0
                    phase: 0.0

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("phase", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class SyncSaw(PureUGen):
    """
    A sawtooth wave that is hard synched to a fundamental pitch.

    ::

        >>> sync_saw = supriya.ugens.SyncSaw.ar(
        ...     saw_frequency=440,
        ...     sync_frequency=440,
        ... )
        >>> sync_saw
        SyncSaw.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("sync_frequency", 440), ("saw_frequency", 440)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class VOsc(PureUGen):
    """
    A wavetable lookup oscillator which can be swept smoothly across wavetables.

    ::

        >>> vosc = supriya.ugens.VOsc.ar(
        ...     buffer_id=supriya.ugens.MouseX.kr(0, 7),
        ...     frequency=440,
        ...     phase=0,
        ... )
        >>> vosc
        VOsc.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("frequency", 440), ("phase", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class VOsc3(PureUGen):
    """
    A wavetable lookup oscillator which can be swept smoothly across wavetables.

    ::

        >>> vosc_3 = supriya.ugens.VOsc3.ar(
        ...     buffer_id=supriya.ugens.MouseX.kr(0, 7),
        ...     freq_1=110,
        ...     freq_2=220,
        ...     freq_3=440,
        ... )
        >>> vosc_3
        VOsc3.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("freq_1", 110), ("freq_2", 220), ("freq_3", 440)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class VarSaw(PureUGen):
    """
    A sawtooth-triangle oscillator with variable duty.

    ::

        >>> supriya.ugens.VarSaw.ar()
        VarSaw.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 440.0), ("initial_phase", 0.0), ("width", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Vibrato(PureUGen):
    """
    Vibrato is a slow frequency modulation.

    ::

        >>> vibrato = supriya.ugens.Vibrato.ar(
        ...     delay=0,
        ...     depth=0.02,
        ...     depth_variation=0.1,
        ...     frequency=440,
        ...     initial_phase=0,
        ...     onset=0,
        ...     rate=6,
        ...     rate_variation=0.04,
        ... )
        >>> vibrato
        Vibrato.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("frequency", 440),
            ("rate", 6),
            ("depth", 0.02),
            ("delay", 0),
            ("onset", 0),
            ("rate_variation", 0.04),
            ("depth_variation", 0.1),
            ("initial_phase", 0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class WrapIndex(UGen):
    """
    A wrapping buffer indexer.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> wrap_index = supriya.ugens.WrapIndex.ar(
        ...     buffer_id=23,
        ...     source=source,
        ... )
        >>> wrap_index
        WrapIndex.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("source", None)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import base64
import hashlib
import pathlib
import pickle
import textwrap
import warnings

import librosa.display
from docutils.nodes import FixedTextElement, General, SkipNode
from matplotlib import pyplot
from uqbar.book.extensions import Extension
from uqbar.strings import normalize

from supriya.ext import websafe_audio
from supriya.io import Player, Plotter


class PlayExtension(Extension):
    template = normalize(
        """
        <audio controls src="{file_path}">
            Your browser does not support the <code>audio</code> element.
        </audio>
        """
    )

    class play_block(General, FixedTextElement):
        pass

    @classmethod
    def setup_console(cls, console, monkeypatch):
        monkeypatch.setattr(
            Player,
            "__call__",
            lambda self: console.push_proxy(cls(self.renderable, self.render_kwargs)),
        )

    @classmethod
    def setup_sphinx(cls, app):
        app.add_node(
            cls.play_block,
            html=[cls.visit_block_html, None],
            latex=[cls.visit_block_latex, None],
            text=[cls.visit_block_text, cls.depart_block_text],
        )

    def __init__(self, renderable, render_kwargs):
        try:
            self.renderable = pickle.loads(pickle.dumps(renderable))
        except TypeError:
            self.renderable = renderable.__render__()
        self.render_kwargs = pickle.loads(pickle.dumps(render_kwargs))

    def to_docutils(self):
        code = "\n".join(
            textwrap.wrap(
                base64.b64encode(
                    pickle.dumps((self.renderable, self.render_kwargs))
                ).decode()
            )
        )
        node = self.play_block(code, code)
        return [node]

    @classmethod
    def render(cls, node, output_path):
        output_path.mkdir(exist_ok=True)
        renderable, render_kwargs = pickle.loads(
            base64.b64decode("".join(node[0].split()))
        )
        return websafe_audio(
            renderable.__render__(render_directory_path=output_path, **render_kwargs)
        )

    @staticmethod
    def visit_block_html(self, node):
        absolute_file_path = PlayExtension.render(
            node, pathlib.Path(self.builder.outdir) / "_images"
        )
        relative_file_path = (
            pathlib.Path(self.builder.imgpath) / absolute_file_path.name
        )
        result = PlayExtension.template.format(file_path=relative_file_path)
        self.body.append(result)
        raise SkipNode


class PlotExtension(Extension):

    template = normalize(
        """
        <object data="{file_path}" type="image/svg+xml"></object>
        """
    )

    class plot_block(General, FixedTextElement):
        pass

    @classmethod
    def setup_console(cls, console, monkeypatch):
        monkeypatch.setattr(
            Plotter, "__call__", lambda self: console.push_proxy(cls(self.plottable))
        )

    @classmethod
    def setup_sphinx(cls, app):
        # Suppress "PySoundFile failed. Trying audioread instead."
        warnings.filterwarnings(
            "ignore", category=UserWarning, module="librosa.core.audio"
        )
        app.add_node(
            cls.plot_block,
            html=[cls.visit_block_html, None],
            latex=[cls.visit_block_latex, None],
            text=[cls.visit_block_text, cls.depart_block_text],
        )

    def __init__(self, plottable, **kwargs):
        self.array, self.sample_rate = plottable.__plot__()

    def to_docutils(self):
        code = "\n".join(
            textwrap.wrap(
                base64.b64encode(pickle.dumps((self.array, self.sample_rate))).decode()
            )
        )
        node = self.plot_block(code, code)
        return [node]

    @classmethod
    def render(cls, node, output_path):
        output_path.mkdir(exist_ok=True)
        array, sample_rate = pickle.loads(base64.b64decode("".join(node[0].split())))
        fig, ax = pyplot.subplots(nrows=1)
        fig.set_size_inches(8, 2)
        fig.patch.set_alpha(0.0)
        ax.set_facecolor("white")
        ax.xaxis.label.set_color("white")
        ax.tick_params(axis="x", colors="white")
        ax.tick_params(axis="y", colors="white")
        librosa.display.waveshow(array, sr=sample_rate, ax=ax)
        hexdigest = hashlib.sha1(node[0].encode()).hexdigest()
        file_path = output_path / f"plot-{hexdigest}.svg"
        fig.savefig(file_path, bbox_inches="tight")
        pyplot.close(fig)
        return file_path

    @staticmethod
    def visit_block_html(self, node):
        absolute_file_path = PlotExtension.render(
            node, pathlib.Path(self.builder.outdir) / "_images"
        )
        relative_file_path = (
            pathlib.Path(self.builder.imgpath) / absolute_file_path.name
        )
        result = PlotExtension.template.format(file_path=relative_file_path)
        self.body.append(result)
        raise SkipNode



from collections.abc import Sequence
from typing import NamedTuple, Tuple

import supriya.osc
from supriya.enums import RequestId

from .bases import Request, Response


class ControlBusFillRequest(Request):
    """
    A /c_fill request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.ControlBusFillRequest(
        ...     index_count_value_triples=[
        ...         (0, 8, 0.5),
        ...         (8, 8, 0.25),
        ...     ],
        ... )
        >>> request
        ControlBusFillRequest(
            index_count_value_triples=(
                (0, 8, 0.5),
                (8, 8, 0.25),
            ),
        )

    ::

        >>> request.to_osc()
        OscMessage('/c_fill', 0, 8, 0.5, 8, 8, 0.25)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.CONTROL_BUS_FILL

    ### INITIALIZER ###

    def __init__(self, index_count_value_triples=None):
        Request.__init__(self)
        if index_count_value_triples:
            triples = []
            for index, count, value in index_count_value_triples:
                index = int(index)
                count = int(count)
                value = float(value)
                assert 0 <= index
                assert 0 < count
                triple = (index, count, value)
                triples.append(triple)
            index_count_value_triples = tuple(triples)
        self._index_count_value_triples = index_count_value_triples

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        if self.index_count_value_triples:
            for index, count, value in self.index_count_value_triples:
                contents.append(index)
                contents.append(count)
                contents.append(value)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def index_count_value_triples(self):
        return self._index_count_value_triples


class ControlBusGetContiguousRequest(Request):
    """
    A /c_getn request.

    ::

        >>> server = supriya.Server().boot()
        >>> request = supriya.commands.ControlBusGetContiguousRequest(
        ...     index_count_pairs=[
        ...         (0, 2),
        ...         (4, 1),
        ...         (8, 2),
        ...         (12, 1),
        ...     ],
        ... )
        >>> request
        ControlBusGetContiguousRequest(
            index_count_pairs=(
                (0, 2),
                (4, 1),
                (8, 2),
                (12, 1),
            ),
        )

    ::

        >>> request.to_osc()
        OscMessage('/c_getn', 0, 2, 4, 1, 8, 2, 12, 1)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...
        ControlBusSetContiguousResponse(
            items=(
                Item(bus_values=(0.0, 0.0), starting_bus_id=0),
                Item(bus_values=(0.0,), starting_bus_id=4),
                Item(bus_values=(0.0, 0.0), starting_bus_id=8),
                Item(bus_values=(0.0,), starting_bus_id=12),
            ),
        )

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/c_getn', 0, 2, 4, 1, 8, 2, 12, 1))
        ('R', OscMessage('/c_setn', 0, 2, 0.0, 0.0, 4, 1, 0.0, 8, 2, 0.0, 0.0, 12, 1, 0.0))

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.CONTROL_BUS_GET_CONTIGUOUS

    ### INITIALIZER ###

    def __init__(self, index_count_pairs=None):
        Request.__init__(self)
        if index_count_pairs:
            pairs = []
            for index, count in index_count_pairs:
                index = int(index)
                count = int(count)
                assert 0 <= index
                assert 0 < count
                pair = (index, count)
                pairs.append(pair)
            index_count_pairs = tuple(pairs)
        self._index_count_pairs = index_count_pairs

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        if self.index_count_pairs:
            for pair in self.index_count_pairs:
                contents.extend(pair)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def index_count_pairs(self):
        return self._index_count_pairs

    @property
    def response_patterns(self):
        return ["/c_setn"], None


class ControlBusGetRequest(Request):
    """
    A /c_get request.

    ::

        >>> server = supriya.Server().boot()
        >>> request = supriya.commands.ControlBusGetRequest(
        ...     indices=(0, 4, 8, 12),
        ... )
        >>> request
        ControlBusGetRequest(
            indices=(0, 4, 8, 12),
        )

    ::

        >>> request.to_osc()
        OscMessage('/c_get', 0, 4, 8, 12)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...
        ControlBusSetResponse(
            items=(
                Item(bus_id=0, bus_value=0.0),
                Item(bus_id=4, bus_value=0.0),
                Item(bus_id=8, bus_value=0.0),
                Item(bus_id=12, bus_value=0.0),
            ),
        )

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/c_get', 0, 4, 8, 12))
        ('R', OscMessage('/c_set', 0, 0.0, 4, 0.0, 8, 0.0, 12, 0.0))

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.CONTROL_BUS_GET

    ### INITIALIZER ###

    def __init__(self, indices=None):
        Request.__init__(self)
        if indices:
            indices = tuple(int(index) for index in indices)
            assert all(0 <= index for index in indices)
        self._indices = indices

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        if self.indices:
            contents.extend(self.indices)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def indices(self):
        return self._indices

    @property
    def response_patterns(self):
        return ["/c_set"], None


class ControlBusSetContiguousRequest(Request):
    """
    A /c_setn request.

    ::

        >>> server = supriya.Server().boot()
        >>> request = supriya.commands.ControlBusSetContiguousRequest(
        ...     index_values_pairs=[
        ...         (0, (0.1, 0.2, 0.3)),
        ...         (4, (0.4, 0.5, 0.6)),
        ...     ],
        ... )
        >>> request
        ControlBusSetContiguousRequest(
            index_values_pairs=(
                (0, (0.1, 0.2, 0.3)),
                (4, (0.4, 0.5, 0.6)),
            ),
        )

    ::

        >>> request.to_osc()
        OscMessage('/c_setn', 0, 3, 0.1, 0.2, 0.3, 4, 3, 0.4, 0.5, 0.6)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...     _ = server.sync()
        ...

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/c_setn', 0, 3, 0.1, 0.2, 0.3, 4, 3, 0.4, 0.5, 0.6))
        ('S', OscMessage('/sync', 0))
        ('R', OscMessage('/synced', 0))

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.CONTROL_BUS_SET_CONTIGUOUS

    ### INITIALIZER ###

    def __init__(self, index_values_pairs=None):
        Request.__init__(self)
        if index_values_pairs:
            pairs = []
            for index, values in index_values_pairs:
                index = int(index)
                values = tuple(float(value) for value in values)
                assert 0 <= index
                assert values
                if not values:
                    continue
                pair = (index, values)
                pairs.append(pair)
            index_values_pairs = tuple(pairs)
        self._index_values_pairs = index_values_pairs

    ### PRIVATE METHODS ###

    def _apply_local(self, server):
        for starting_bus_index, values in self._index_values_pairs or ():
            for i, value in enumerate(values):
                bus_id = starting_bus_index + i
                bus_proxy = server._get_control_bus_proxy(bus_id)
                bus_proxy._value = value

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        if self.index_values_pairs:
            for index, values in self.index_values_pairs:
                contents.append(index)
                contents.append(len(values))
                contents.extend(values)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def index_values_pairs(self):
        return self._index_values_pairs


class ControlBusSetContiguousResponse(Response, Sequence):

    ### CLASS VARIABLES ###

    class Item(NamedTuple):
        bus_values: Tuple[float]
        starting_bus_id: int

    ### INITIALIZER ###

    def __init__(self, items=None):
        self._items = items

    ### SPECIAL METHODS ###

    def __getitem__(self, item):
        return self._items[item]

    def __len__(self):
        return len(self._items)

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        items = []
        contents = list(osc_message.contents)
        while contents:
            starting_bus_id = contents[0]
            bus_count = contents[1]
            bus_values = tuple(contents[2 : 2 + bus_count])
            item = cls.Item(starting_bus_id=starting_bus_id, bus_values=bus_values)
            items.append(item)
            contents = contents[2 + bus_count :]
        items = tuple(items)
        response = cls(items=items)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def items(self):
        return self._items


class ControlBusSetRequest(Request):
    """
    A /c_set request.

    ::

        >>> server = supriya.Server().boot()
        >>> request = supriya.commands.ControlBusSetRequest(
        ...     index_value_pairs=[
        ...         (0, 0.1),
        ...         (1, 0.2),
        ...         (2, 0.3),
        ...         (3, 0.4),
        ...     ],
        ... )
        >>> request
        ControlBusSetRequest(
            index_value_pairs=(
                (0, 0.1),
                (1, 0.2),
                (2, 0.3),
                (3, 0.4),
            ),
        )

    ::

        >>> request.to_osc()
        OscMessage('/c_set', 0, 0.1, 1, 0.2, 2, 0.3, 3, 0.4)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...     _ = server.sync()
        ...

    ::

        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/c_set', 0, 0.1, 1, 0.2, 2, 0.3, 3, 0.4))
        ('S', OscMessage('/sync', 0))
        ('R', OscMessage('/synced', 0))


    """

    ### CLASS VARIABLES ###

    request_id = RequestId.CONTROL_BUS_SET

    ### INITIALIZER ###

    def __init__(self, index_value_pairs=None):
        Request.__init__(self)
        if index_value_pairs:
            pairs = []
            for index, value in index_value_pairs:
                index = int(index)
                value = float(value)
                assert 0 <= index
                pair = (index, value)
                pairs.append(pair)
            index_value_pairs = tuple(pairs)
        self._index_value_pairs = index_value_pairs

    ### PRIVATE METHODS ###

    def _apply_local(self, server):
        for bus_id, value in self.index_value_pairs or ():
            bus_proxy = server._get_control_bus_proxy(bus_id)
            bus_proxy._value = value

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        if self.index_value_pairs:
            for pair in self.index_value_pairs:
                contents.extend(pair)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def index_value_pairs(self):
        return self._index_value_pairs


class ControlBusSetResponse(Response, Sequence):

    ### CLASS VARIABLES ###

    class Item(NamedTuple):
        bus_id: int
        bus_value: float

    ### INITIALIZER ###

    def __init__(self, items=None):
        self._items = items

    ### SPECIAL METHODS ###

    def __getitem__(self, item):
        return self._items[item]

    def __len__(self):
        return len(self._items)

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        items = []
        for group in cls._group_items(osc_message.contents, 2):
            item = cls.Item(*group)
            items.append(item)
        response = cls(items=tuple(items))
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def items(self):
        return self._items



from uqbar.objects import get_repr

from supriya.system import SupriyaObject

from .IntervalTreeDriver import IntervalTreeDriver
from .Moment import Moment


class IntervalTree(SupriyaObject):
    """
    A mutable always-sorted collection of intervals.

    ::

        >>> from supriya.intervals import Interval, IntervalTree
        >>> intervals = (
        ...     Interval(0, 3),
        ...     Interval(1, 3),
        ...     Interval(1, 2),
        ...     Interval(2, 5),
        ...     Interval(6, 9),
        ... )
        >>> interval_tree = IntervalTree(intervals)

    """

    ### INITIALIZER ###

    # TODO: Protect datastructure with a R/W lock: https://pypi.org/project/readerwriterlock/

    def __init__(self, intervals=None, accelerated=True):
        self._driver = IntervalTreeDriver(intervals)
        self._accelerated = bool(accelerated)
        if not accelerated:
            return
        try:
            import pyximport  # noqa

            from .IntervalTreeDriverEx import IntervalTreeDriverEx

            self._driver = IntervalTreeDriverEx(intervals)
        except (ImportError, ModuleNotFoundError):
            pass

    ### SPECIAL METHODS ###

    def __and__(self, interval):
        new_intervals = []
        for current_interval in self[:]:
            result = current_interval & interval
            new_intervals.extend(result)
        self[:] = sorted(new_intervals)
        return self

    def __contains__(self, interval):
        """
        Is true if this interval tree contains `interval`. Otherwise
        false.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> intervals[0] in interval_tree
            True

        ::

            >>> Interval(-1, 100) in interval_tree
            False

        Returns boolean.
        """
        return interval in self._driver

    def __getitem__(self, item):
        """
        Gets interval at index `item`.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> interval_tree[-1]
            Interval(start_offset=6.0, stop_offset=9.0)

        ::

            >>> for interval in interval_tree[:3]:
            ...     interval
            ...
            Interval(start_offset=0.0, stop_offset=3.0)
            Interval(start_offset=1.0, stop_offset=2.0)
            Interval(start_offset=1.0, stop_offset=3.0)

        Returns interval or intervals.
        """
        return self._driver[item]

    def __getstate__(self):
        return self._accelerated, tuple(self)

    def __iter__(self):
        """
        Iterates intervals in this interval tree.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> for interval in interval_tree:
            ...     interval
            ...
            Interval(start_offset=0.0, stop_offset=3.0)
            Interval(start_offset=1.0, stop_offset=2.0)
            Interval(start_offset=1.0, stop_offset=3.0)
            Interval(start_offset=2.0, stop_offset=5.0)
            Interval(start_offset=6.0, stop_offset=9.0)

        Returns generator.
        """
        return iter(self._driver)

    def __len__(self):
        """
        Gets length of this interval tree.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> len(interval_tree)
            5

        Returns integer.
        """
        return len(self._driver)

    def __repr__(self):
        if not len(self):
            return get_repr(self, multiline=False)
        return get_repr(self, multiline=True)

    def __setitem__(self, i, new):
        """
        Sets intervals at index `i` to `new`.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> interval_tree[:3] = [Interval(100, 200)]

        Returns none.
        """
        if isinstance(i, int):
            self.remove(self[i])
            self.add(new)
        elif isinstance(i, slice):
            for interval in self[i]:
                self.remove(interval)
            self.update(new)
        else:
            message = "Indices must be ints or slices, got {}".format(i)
            raise TypeError(message)

    def __setstate__(self, state):
        accelerated, intervals = state
        self.__init__(intervals=intervals, accelerated=accelerated)

    def __sub__(self, interval):
        """
        Delete material that intersects `interval`:

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> interval_tree = IntervalTree(
            ...     [
            ...         Interval(0, 16),
            ...         Interval(5, 12),
            ...         Interval(-2, 8),
            ...     ]
            ... )

        ::

            >>> interval = Interval(5, 10)
            >>> result = interval_tree - interval

        ::

            >>> for interval in interval_tree:
            ...     interval
            ...
            Interval(start_offset=-2.0, stop_offset=5.0)
            Interval(start_offset=0.0, stop_offset=5.0)
            Interval(start_offset=10.0, stop_offset=12.0)
            Interval(start_offset=10.0, stop_offset=16.0)

        Operates in place and returns interval tree.
        """
        if not self._is_interval(interval):
            raise ValueError(interval)
        intersection = self.find_intersection(interval)
        to_update = []
        for intersecting_interval in intersection:
            self.remove(intersecting_interval)
            for x in intersecting_interval - interval:
                to_update.append(x)
        self.update(to_update)
        return self

    ### PRIVATE METHODS ###

    @staticmethod
    def _is_interval(expr):
        if hasattr(expr, "start_offset") and hasattr(expr, "stop_offset"):
            return True
        return False

    ### PUBLIC METHODS ###

    def find_intersection(self, interval_or_offset):
        """
        Find intervals intersecting a interval or offset.

        ..  container:: example

            Finds intervals overlapping `offset`.

            ::

                >>> from supriya.intervals import Interval, IntervalTree
                >>> intervals = (
                ...     Interval(0, 3),
                ...     Interval(1, 3),
                ...     Interval(1, 2),
                ...     Interval(2, 5),
                ...     Interval(6, 9),
                ... )
                >>> interval_tree = IntervalTree(intervals)

            ::

                >>> for x in interval_tree.find_intersection(1.5):
                ...     x
                ...
                Interval(start_offset=0.0, stop_offset=3.0)
                Interval(start_offset=1.0, stop_offset=2.0)
                Interval(start_offset=1.0, stop_offset=3.0)

        ..  container:: example

            Finds intervals overlapping `interval`.

            ::

                >>> intervals = (
                ...     Interval(0, 3),
                ...     Interval(1, 3),
                ...     Interval(1, 2),
                ...     Interval(2, 5),
                ...     Interval(6, 9),
                ... )
                >>> interval_tree = IntervalTree(intervals)

            ::

                >>> interval = Interval(2, 4)
                >>> for x in interval_tree.find_intersection(interval):
                ...     x
                ...
                Interval(start_offset=0.0, stop_offset=3.0)
                Interval(start_offset=1.0, stop_offset=3.0)
                Interval(start_offset=2.0, stop_offset=5.0)

        """
        if self._is_interval(interval_or_offset):
            return self._driver.find_intervals_intersecting_interval(interval_or_offset)
        return self._driver.find_intervals_intersecting_offset(interval_or_offset)

    def find_intervals_starting_at(self, offset):
        return self._driver.find_intervals_starting_at(offset)

    def find_intervals_stopping_at(self, offset):
        return self._driver.find_intervals_stopping_at(offset)

    def get_moment_at(self, offset):
        """
        Gets moment at `offset`.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> interval_tree.get_moment_at(1)
            <Moment(1 <<3>>)>

        ::

            >>> interval_tree.get_moment_at(6.5)
            <Moment(6.5 <<1>>)>

        """
        stop_intervals = self.find_intervals_stopping_at(offset)
        start_intervals, overlap_intervals = [], []
        for interval in self.find_intersection(offset):
            if interval.start_offset == offset:
                start_intervals.append(interval)
            else:
                overlap_intervals.append(interval)
        moment = Moment(
            interval_tree=self,
            overlap_intervals=overlap_intervals,
            start_intervals=start_intervals,
            start_offset=offset,
            stop_intervals=stop_intervals,
        )
        return moment

    def get_offset_after(self, offset):
        """
        Gets first start or stop offset after ``offset``, otherwise None.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> for i in range(-1, 11):
            ...     print(i, interval_tree.get_offset_after(i))
            ...
            -1 0.0
            0 1.0
            1 2.0
            2 3.0
            3 5.0
            4 5.0
            5 6.0
            6 9.0
            7 9.0
            8 9.0
            9 None
            10 None

        """
        return self._driver.get_offset_after(float(offset))

    def get_start_offset_after(self, offset):
        """
        Gets start offst in this interval tree after `offset`.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> interval_tree.get_start_offset_after(-1)
            0.0

        ::

            >>> interval_tree.get_start_offset_after(0)
            1.0

        ::

            >>> interval_tree.get_start_offset_after(1)
            2.0

        ::

            >>> interval_tree.get_start_offset_after(2)
            6.0

        ::

            >>> interval_tree.get_start_offset_after(6) is None
            True

        """
        return self._driver.get_start_offset_after(float(offset))

    def get_start_offset_before(self, offset):
        """
        Gets start offst in this interval tree before `offset`.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> interval_tree.get_start_offset_before(7)
            6.0

        ::

            >>> interval_tree.get_start_offset_before(6)
            2.0

        ::

            >>> interval_tree.get_start_offset_before(2)
            1.0

        ::

            >>> interval_tree.get_start_offset_before(1)
            0.0

        ::

            >>> interval_tree.get_start_offset_before(0) is None
            True

        """
        return self._driver.get_start_offset_before(float(offset))

    def index(self, interval):
        return self._driver.index(interval)

    def add(self, interval):
        self._driver.add(interval)

    def update(self, intervals):
        self._driver.update(intervals)

    def iterate_moments(self, reverse=False):
        """
        Iterates moments in this interval tree.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> for x in interval_tree.iterate_moments():
            ...     x
            ...
            <Moment(0.0 <<1>>)>
            <Moment(1.0 <<3>>)>
            <Moment(2.0 <<3>>)>
            <Moment(6.0 <<1>>)>

        ::

            >>> for x in interval_tree.iterate_moments(reverse=True):
            ...     x
            ...
            <Moment(6.0 <<1>>)>
            <Moment(2.0 <<3>>)>
            <Moment(1.0 <<3>>)>
            <Moment(0.0 <<1>>)>

        Returns generator.
        """

        if reverse:
            start_offset = self.latest_start_offset
            moment = self.get_moment_at(start_offset)
            yield moment
            moment = moment.previous_moment
            while moment is not None:
                yield moment
                moment = moment.previous_moment
        else:
            start_offset = self.earliest_start_offset
            moment = self.get_moment_at(start_offset)
            yield moment
            moment = moment.next_moment
            while moment is not None:
                yield moment
                moment = moment.next_moment

    def iterate_moments_nwise(self, n=3, reverse=False):
        """
        Iterates moments in this interval tree in groups of
        `n`.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> for x in interval_tree.iterate_moments_nwise(n=2):
            ...     x
            ...
            [<Moment(0.0 <<1>>)>, <Moment(1.0 <<3>>)>]
            [<Moment(1.0 <<3>>)>, <Moment(2.0 <<3>>)>]
            [<Moment(2.0 <<3>>)>, <Moment(6.0 <<1>>)>]

        ::

            >>> for x in interval_tree.iterate_moments_nwise(n=2, reverse=True):
            ...     x
            ...
            [<Moment(2.0 <<3>>)>, <Moment(6.0 <<1>>)>]
            [<Moment(1.0 <<3>>)>, <Moment(2.0 <<3>>)>]
            [<Moment(0.0 <<1>>)>, <Moment(1.0 <<3>>)>]

        Returns generator.
        """
        n = int(n)
        assert 0 < n
        if reverse:
            for moment in self.iterate_moments(reverse=True):
                moments = [moment]
                while len(moments) < n:
                    next_moment = moments[-1].next_moment
                    if next_moment is None:
                        break
                    moments.append(next_moment)
                if len(moments) == n:
                    yield moments
        else:
            for moment in self.iterate_moments():
                moments = [moment]
                while len(moments) < n:
                    previous_moment = moments[-1].previous_moment
                    if previous_moment is None:
                        break
                    moments.append(previous_moment)
                if len(moments) == n:
                    yield list(reversed(moments))

    def remove(self, interval):
        """
        Removes interval from this interval tree.

        ::

            >>> from supriya.intervals import Interval, IntervalTree
            >>> intervals = (
            ...     Interval(0, 3),
            ...     Interval(1, 3),
            ...     Interval(1, 2),
            ...     Interval(2, 5),
            ...     Interval(6, 9),
            ... )
            >>> interval_tree = IntervalTree(intervals)

        ::

            >>> for interval in intervals[1:-1]:
            ...     interval_tree.remove(interval)
            ...

        ::

            >>> for interval in interval_tree:
            ...     interval
            ...
            Interval(start_offset=0.0, stop_offset=3.0)
            Interval(start_offset=6.0, stop_offset=9.0)

        """
        self._driver.remove(interval)

    ### PRIVATE PROPERTIES ###

    @property
    def _root_node(self):
        return self._driver._root_node

    @_root_node.setter
    def _root_node(self, node):
        self._driver._root_node = node

    ### PUBLIC PROPERTIES ###

    @property
    def all_offsets(self):
        offsets = set()
        for interval in self:
            offsets.add(interval.start_offset)
            offsets.add(interval.stop_offset)
        return tuple(sorted(offsets))

    @property
    def all_start_offsets(self):
        start_offsets = set()
        for interval in self:
            start_offsets.add(interval.start_offset)
        return tuple(sorted(start_offsets))

    @property
    def all_stop_offsets(self):
        stop_offsets = set()
        for interval in self:
            stop_offsets.add(interval.stop_offset)
        return tuple(sorted(stop_offsets))

    @property
    def earliest_start_offset(self):
        def recurse(node):
            if node.left_child is not None:
                return recurse(node.left_child)
            return node.start_offset

        if self._root_node is not None:
            return recurse(self._root_node)
        return float("-inf")

    @property
    def earliest_stop_offset(self):
        if self._root_node is not None:
            return self._root_node.stop_offset_low
        return float("inf")

    @property
    def intervals(self):
        return list(self)

    @property
    def latest_start_offset(self):
        def recurse(node):
            if node.right_child is not None:
                return recurse(node.right_child)
            return node.start_offset

        if self._root_node is not None:
            return recurse(self._root_node)
        return float("-inf")

    @property
    def latest_stop_offset(self):
        if self._root_node is not None:
            return self._root_node.stop_offset_high
        return float("inf")

    @property
    def start_offset(self):
        return self.earliest_start_offset

    @property
    def stop_offset(self):
        return self.latest_stop_offset



from os import PathLike
from pathlib import Path
from typing import Dict, Optional, SupportsFloat, SupportsInt, Union

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

from .enums import AddAction, CalculationRate, HeaderFormat, SampleFormat


class Default:
    pass


class Missing:
    pass


class SupportsRender(Protocol):
    def __render__(
        self,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        **kwargs,
    ) -> Union[Path, "SupportsRender"]:
        ...


AddActionLike = Optional[Union[AddAction, SupportsInt, str]]
CalculationRateLike = Optional[Union[CalculationRate, SupportsInt, str]]
HeaderFormatLike = Optional[Union[HeaderFormat, SupportsInt, str]]
SampleFormatLike = Optional[Union[SampleFormat, SupportsInt, str]]
UGenInputMap = Optional[Dict[str, Union[SupportsFloat, str, None]]]



import collections

from supriya import CalculationRate
from supriya.synthdefs import PureUGen, UGen


class AllpassC(PureUGen):
    """
    A cubic-interpolating allpass delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> allpass_c = supriya.ugens.AllpassC.ar(source=source)
        >>> allpass_c
        AllpassC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class AllpassL(PureUGen):
    """
    A linear interpolating allpass delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> allpass_l = supriya.ugens.AllpassL.ar(source=source)
        >>> allpass_l
        AllpassL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class AllpassN(PureUGen):
    """
    A non-interpolating allpass delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> allpass_n = supriya.ugens.AllpassN.ar(source=source)
        >>> allpass_n
        AllpassN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BufAllpassC(PureUGen):
    """
    A buffer-based cubic-interpolating allpass delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufAllpassC.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufAllpassC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BufAllpassL(PureUGen):
    """
    A buffer-based linear-interpolating allpass delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufAllpassL.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufAllpassL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BufAllpassN(PureUGen):
    """
    A buffer-based non-interpolating allpass delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufAllpassN.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufAllpassN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BufCombC(PureUGen):
    """
    A buffer-based cubic-interpolating comb delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufCombC.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufCombC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.AUDIO)


class BufCombL(PureUGen):
    """
    A buffer-based linear-interpolating comb delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufCombL.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufCombL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.AUDIO)


class BufCombN(PureUGen):
    """
    A buffer-based non-interpolating comb delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufCombN.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufCombN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.AUDIO)


class BufDelayC(PureUGen):
    """
    A buffer-based cubic-interpolating delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufDelayC.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufDelayC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BufDelayL(PureUGen):
    """
    A buffer-based linear-interpolating delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufDelayL.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufDelayL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BufDelayN(PureUGen):
    """
    A buffer-based non-interpolating delay line unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.BufDelayN.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )
        BufDelayN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class CombC(PureUGen):
    """
    A cubic-interpolating comb delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.CombC.ar(source=source)
        CombC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class CombL(PureUGen):
    """
    A linear interpolating comb delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.CombL.ar(source=source)
        CombL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class CombN(PureUGen):
    """
    A non-interpolating comb delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.CombN.ar(source=source)
        CombN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("maximum_delay_time", 0.2),
            ("delay_time", 0.2),
            ("decay_time", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class DelTapRd(UGen):
    """
    A delay tap reader unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.SoundIn.ar(0)
        >>> tapin = supriya.ugens.DelTapWr.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )

    ::

        >>> tapin
        DelTapWr.ar()

    ::

        >>> tapout = supriya.ugens.DelTapRd.ar(
        ...     buffer_id=buffer_id,
        ...     phase=tapin,
        ...     delay_time=0.1,
        ...     interpolation=True,
        ... )

    ::

        >>> tapout
        DelTapRd.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("buffer_id", None),
            ("phase", None),
            ("delay_time", 0.0),
            ("interpolation", 1.0),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class DelTapWr(UGen):
    """
    A delay tap writer unit generator.

    ::

        >>> buffer_id = 0
        >>> source = supriya.ugens.SoundIn.ar(0)
        >>> tapin = supriya.ugens.DelTapWr.ar(
        ...     buffer_id=buffer_id,
        ...     source=source,
        ... )

    ::

        >>> tapin
        DelTapWr.ar()

    ::

        >>> tapout = supriya.ugens.DelTapRd.ar(
        ...     buffer_id=buffer_id,
        ...     phase=tapin,
        ...     delay_time=0.1,
        ...     interpolation=True,
        ... )

    ::

        >>> tapout
        DelTapRd.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("buffer_id", None), ("source", None)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class DelayC(PureUGen):
    """
    A cubic-interpolating delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.DelayC.ar(source=source)
        DelayC.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("maximum_delay_time", 0.2), ("delay_time", 0.2)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class DelayL(PureUGen):
    """
    A linear-interpolating delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.DelayL.ar(source=source)
        DelayL.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("maximum_delay_time", 0.2), ("delay_time", 0.2)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class DelayN(PureUGen):
    """
    A non-interpolating delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.DelayN.ar(source=source)
        DelayN.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("maximum_delay_time", 0.2), ("delay_time", 0.2)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Delay1(PureUGen):
    """
    A one-sample delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.Delay1.ar(source=source)
        Delay1.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Delay2(PureUGen):
    """
    A two-sample delay line unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.Delay2.ar(source=source)
        Delay2.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("maximum_delay_time", 0.2), ("delay_time", 0.2)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import configparser
import logging
import pathlib

import platformdirs

try:
    import pyximport

    pyximport.install(language_level=3)
    del pyximport
except ImportError:
    pass

output_path = pathlib.Path(platformdirs.user_cache_dir("supriya", "supriya"))
if not output_path.exists():
    try:
        output_path.mkdir(parents=True, exist_ok=True)
    except IOError:
        pass

config = configparser.ConfigParser()
config.read_dict({"core": {"scsynth_path": "scsynth"}})
config_path = pathlib.Path(platformdirs.user_config_dir("supriya", "supriya"))
config_path = config_path / "supriya.cfg"
if not config_path.exists():
    try:
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with config_path.open("w") as file_pointer:
            config.write(file_pointer, True)
    except IOError:
        pass
with config_path.open() as file_pointer:
    config.read_file(file_pointer)

del platformdirs
del configparser
del file_pointer
del pathlib


def _setup_logging(*loggers):
    logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    for logger in loggers:
        logging.getLogger(logger).setLevel(logging.DEBUG)


from ._version import __version__, __version_info__  # noqa
from . import utils  # noqa
from .clocks import AsyncClock, AsyncOfflineClock, Clock, OfflineClock  # noqa
from .enums import (  # noqa
    AddAction,
    BinaryOperator,
    CalculationRate,
    DoneAction,
    EnvelopeShape,
    HeaderFormat,
    NodeAction,
    ParameterRate,
    RequestId,
    RequestName,
    SampleFormat,
    SignalRange,
    UnaryOperator,
    Unit,
)
from .io import graph, play, plot, render  # noqa
from .synthdefs import (  # noqa
    Envelope,
    Parameter,
    Range,
    SynthDef,
    SynthDefBuilder,
    SynthDefFactory,
)
from .realtime import (  # noqa
    AsyncServer,
    Buffer,
    BufferGroup,
    Bus,
    BusGroup,
    Group,
    Synth,
    Server,
)
from . import assets  # noqa
from .nonrealtime import Session  # noqa
from .providers import Provider  # noqa
from .scsynth import Options  # noqa
from .soundfiles import Say, SoundFile  # noqa
from .system import Assets  # noqa



import supriya.synthdefs
import supriya.ugens
from supriya import ParameterRate


def _build_test_synthdef():
    with supriya.synthdefs.SynthDefBuilder(
        frequency=440,
        amplitude=supriya.synthdefs.Parameter(
            value=1.0, parameter_rate=ParameterRate.AUDIO
        ),
    ) as builder:
        sin_osc = supriya.ugens.SinOsc.ar(frequency=builder["frequency"])
        enveloped_sin = sin_osc * builder["amplitude"]
        supriya.ugens.Out.ar(bus=0, source=enveloped_sin)
    synthdef = builder.build(name="test")
    return synthdef


test = _build_test_synthdef()

__all__ = ("test",)



import abc
import collections

from supriya import CalculationRate
from supriya.synthdefs import PseudoUGen, PureUGen, UGen


class A2K(PureUGen):
    """
    An audio-rate to control-rate convert unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> a_2_k = supriya.ugens.A2K.kr(
        ...     source=source,
        ... )
        >>> a_2_k
        A2K.kr()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.CONTROL,)


class AmpComp(PureUGen):
    """
    Basic psychoacoustic amplitude compensation.

    ::

        >>> amp_comp = supriya.ugens.AmpComp.ar(
        ...     exp=0.3333,
        ...     frequency=1000,
        ...     root=0,
        ... )
        >>> amp_comp
        AmpComp.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 1000), ("root", 0), ("exp", 0.3333)]
    )
    _valid_calculation_rates = (
        CalculationRate.AUDIO,
        CalculationRate.CONTROL,
        CalculationRate.SCALAR,
    )


class AmpCompA(PureUGen):
    """
    Basic psychoacoustic amplitude compensation (ANSI A-weighting curve).

    ::

        >>> amp_comp_a = supriya.ugens.AmpCompA.ar(
        ...     frequency=1000,
        ...     min_amp=0.32,
        ...     root=0,
        ...     root_amp=1,
        ... )
        >>> amp_comp_a
        AmpCompA.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("frequency", 1000), ("root", 0), ("min_amp", 0.32), ("root_amp", 1)]
    )
    _valid_calculation_rates = (
        CalculationRate.AUDIO,
        CalculationRate.CONTROL,
        CalculationRate.SCALAR,
    )


class DC(PureUGen):
    """
    A DC unit generator.

    ::

        >>> supriya.ugens.DC.ar(
        ...     source=0,
        ... )
        DC.ar()

    ::

        >>> supriya.ugens.DC.ar(
        ...     source=(1, 2, 3),
        ... )
        UGenArray({3})

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class K2A(PureUGen):
    """
    A control-rate to audio-rate converter unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.kr()
        >>> k_2_a = supriya.ugens.K2A.ar(
        ...     source=source,
        ... )
        >>> k_2_a
        K2A.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class LinExp(PureUGen):
    """
    A linear-to-exponential range mapper.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> lin_exp = supriya.ugens.LinExp.ar(
        ...     input_maximum=1.0,
        ...     input_minimum=-1.0,
        ...     output_maximum=22050,
        ...     output_minimum=20,
        ...     source=source,
        ... )
        >>> lin_exp
        LinExp.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("input_minimum", 0),
            ("input_maximum", 1),
            ("output_minimum", 1),
            ("output_maximum", 2),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LinLin(PseudoUGen):

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self):
        raise NotImplementedError

    ### PUBLIC METHODS ###

    @staticmethod
    def ar(
        source=None,
        input_minimum=0.0,
        input_maximum=1.0,
        output_minimum=1.0,
        output_maximum=2.0,
    ):
        import supriya.ugens

        scale = (output_maximum - output_minimum) / (input_maximum - input_minimum)
        offset = output_minimum - (scale * input_minimum)
        ugen = supriya.ugens.MulAdd.new(source=source, multiplier=scale, addend=offset)
        return ugen

    @staticmethod
    def kr(
        source=None,
        input_minimum=0.0,
        input_maximum=1.0,
        output_minimum=1.0,
        output_maximum=2.0,
    ):
        import supriya.ugens

        scale = (output_maximum - output_minimum) / (input_maximum - input_minimum)
        offset = output_minimum - (scale * input_minimum)
        ugen = supriya.ugens.MulAdd.new(source=source, multiplier=scale, addend=offset)
        return ugen


class Line(UGen):
    """
    A line generating unit generator.

    ::

        >>> supriya.ugens.Line.ar()
        Line.ar()

    """

    ### CLASS VARIABLES ###

    _has_done_flag = True
    _ordered_input_names = collections.OrderedDict(
        [("start", 0.0), ("stop", 1.0), ("duration", 1.0), ("done_action", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### PRIVATE METHODS ###

    @classmethod
    def _new_expanded(
        cls,
        calculation_rate=None,
        done_action=None,
        duration=None,
        stop=None,
        start=None,
    ):
        import supriya.synthdefs

        done_action = supriya.DoneAction.from_expr(int(done_action))
        return super(Line, cls)._new_expanded(
            calculation_rate=calculation_rate,
            done_action=done_action,
            duration=duration,
            stop=stop,
            start=start,
        )


class Silence(PseudoUGen):
    """
    An audio-rate silence pseudo-unit generator.

    ::

        >>> supriya.ugens.Silence.ar(channel_count=2)
        UGenArray({2})

    """

    @classmethod
    def ar(cls, channel_count=1):
        import supriya.synthdefs
        import supriya.ugens

        channel_count = int(channel_count)
        assert 0 <= channel_count
        silence = supriya.ugens.DC.ar(0)
        if channel_count == 1:
            return silence
        output_proxies = [silence[0]] * channel_count
        return supriya.synthdefs.UGenArray(output_proxies)


class XLine(UGen):
    """
    An exponential line generating unit generator.

    ::

        >>> supriya.ugens.XLine.ar()
        XLine.ar()

    """

    _has_done_flag = True
    _ordered_input_names = collections.OrderedDict(
        [("start", 0.0), ("stop", 1.0), ("duration", 1.0), ("done_action", 0.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import collections
from collections.abc import Sequence

from supriya import CalculationRate
from supriya.synthdefs import UGen


class Clip(UGen):
    """
    Clips a signal outside given thresholds.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> clip = supriya.ugens.Clip.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> clip
        Clip.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("minimum", 0.0), ("maximum", 1.0)]
    )
    _valid_calculation_rates = (
        CalculationRate.AUDIO,
        CalculationRate.CONTROL,
        CalculationRate.SCALAR,
    )


class Fold(UGen):
    """
    Folds a signal outside given thresholds.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> fold = supriya.ugens.Fold.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> fold
        Fold.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("minimum", 0.0), ("maximum", 1.0)]
    )
    _valid_calculation_rates = (
        CalculationRate.AUDIO,
        CalculationRate.CONTROL,
        CalculationRate.SCALAR,
    )


class Gate(UGen):
    """
    Gates or holds.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> trigger = supriya.ugens.Dust.kr(1)
        >>> gate = supriya.ugens.Gate.ar(
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> gate
        Gate.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None), ("trigger", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class InRange(UGen):
    """
    Tests if a signal is within a given range.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> in_range = supriya.ugens.InRange.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> in_range
        InRange.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", 0), ("minimum", 0), ("maximum", 1)]
    )
    _valid_calculation_rates = (
        CalculationRate.AUDIO,
        CalculationRate.CONTROL,
        CalculationRate.SCALAR,
    )


class Latch(UGen):
    """
    Samples and holds.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> trigger = supriya.ugens.Dust.kr(1)
        >>> latch = supriya.ugens.Latch.ar(
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> latch
        Latch.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None), ("trigger", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LeastChange(UGen):
    """
    Outputs least changed input.

    ::

        >>> least_change = supriya.ugens.LeastChange.ar(
        ...     a=0,
        ...     b=0,
        ... )
        >>> least_change
        LeastChange.ar()

    """

    _ordered_input_names = collections.OrderedDict([("a", 0), ("b", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class MostChange(UGen):
    """
    Outputs most changed input.

    ::

        >>> most_change = supriya.ugens.MostChange.ar(
        ...     a=0,
        ...     b=0,
        ... )
        >>> most_change
        MostChange.ar()

    """

    _ordered_input_names = collections.OrderedDict([("a", 0), ("b", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Peak(UGen):
    """
    Tracks peak signal amplitude.

    ::

        >>> source = supriya.ugens.In.ar(0)
        >>> trigger = supriya.ugens.Impulse.kr(1)
        >>> peak = supriya.ugens.Peak.ar(
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> peak
        Peak.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None), ("trigger", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class PeakFollower(UGen):
    """
    Tracks peak signal amplitude.

    ::

        >>> source = supriya.ugens.In.ar(0)
        >>> peak_follower = supriya.ugens.PeakFollower.ar(
        ...     decay=0.999,
        ...     source=source,
        ... )
        >>> peak_follower
        PeakFollower.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None), ("decay", 0.999)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Phasor(UGen):
    """
    A resettable linear ramp between two levels.

    ::

        >>> trigger = supriya.ugens.Impulse.kr(0.5)
        >>> phasor = supriya.ugens.Phasor.ar(
        ...     rate=1,
        ...     reset_pos=0,
        ...     start=0,
        ...     stop=1,
        ...     trigger=trigger,
        ... )
        >>> phasor
        Phasor.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("trigger", 0), ("rate", 1), ("start", 0), ("stop", 1), ("reset_pos", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Poll(UGen):
    """
    A UGen poller.

    ::

        >>> sine = supriya.ugens.SinOsc.ar()
        >>> trigger = supriya.ugens.Impulse.kr(1)
        >>> poll = supriya.ugens.Poll.ar(
        ...     source=sine,
        ...     trigger=trigger,
        ...     trigger_id=1234,
        ... )
        >>> poll
        Poll.ar()

    ..  container:: example

        Unlike **sclang**, Python does not share any inter-process
        communication with **scsynth**. This means that the Poll UGen is not
        able to automatically print out its diagnostic messages into a Python
        interpreter session.

        To get information out of the Poll UGen, we first need to set the
        Poll's `trigger_id` to a value greater than 0. This will cause the poll
        to send `/tr` OSC messages back to its client - Python. We can then
        register a callback to respond to these `/tr` messages.

        ::

            >>> with supriya.SynthDefBuilder() as builder:
            ...     sine = supriya.ugens.SinOsc.ar()
            ...     trigger = supriya.ugens.Impulse.kr(1)
            ...     poll = supriya.ugens.Poll.ar(
            ...         source=sine,
            ...         trigger=trigger,
            ...         trigger_id=1234,
            ...     )
            ...
            >>> synthdef = builder.build()

        ::

            >>> server = supriya.Server().boot()
            >>> synth = supriya.Synth(synthdef).allocate(server)
            >>> callback = server.osc_protocol.register(
            ...     pattern="/tr",
            ...     procedure=lambda response: print(
            ...         "Poll value is: {}".format(response.value)
            ...     ),
            ...     once=True,
            ... )

        ::

            >>> server.quit()
            <Server: offline>

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [("trigger", None), ("source", None), ("trigger_id", -1)]
    )

    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### INITIALIZER ###

    def __init__(
        self,
        calculation_rate=None,
        label=None,
        source=None,
        trigger=None,
        trigger_id=-1,
    ):
        import supriya.synthdefs
        import supriya.ugens

        if label is None:
            if isinstance(source, supriya.synthdefs.UGen):
                label = type(source).__name__
            elif isinstance(source, supriya.synthdefs.OutputProxy):
                label = type(source.source).__name__
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )
        label = str(label)
        self._configure_input("label", len(label))
        for character in label:
            self._configure_input("label", ord(character))

    ### PUBLIC METHODS ###

    @classmethod
    def ar(cls, label=None, source=None, trigger=None, trigger_id=-1):
        import supriya.synthdefs

        calculation_rate = supriya.CalculationRate.AUDIO
        ugen = cls._new_expanded(
            calculation_rate=calculation_rate,
            label=label,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )
        return ugen

    @classmethod
    def kr(cls, label=None, source=None, trigger=None, trigger_id=-1):
        import supriya.synthdefs

        calculation_rate = supriya.CalculationRate.CONTROL
        ugen = cls._new_expanded(
            calculation_rate=calculation_rate,
            label=label,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )
        return ugen

    @classmethod
    def new(cls, label=None, source=None, trigger=None, trigger_id=-1):
        import supriya.synthdefs

        if isinstance(source, Sequence):
            source = (source,)
        calculation_rates = []
        for single_source in source:
            rate = supriya.CalculationRate.from_expr(single_source)
            calculation_rates.append(rate)
        ugen = cls._new_expanded(
            calculation_rate=calculation_rates,
            label=label,
            source=source,
            trigger=trigger,
            trigger_id=trigger_id,
        )
        return ugen

    ### PUBLIC PROPERTIES ###

    @property
    def label(self):
        """
        Gets `label` input of Poll.

        ::

            >>> sine = supriya.ugens.SinOsc.ar()
            >>> trigger = supriya.ugens.Impulse.kr(1)
            >>> poll = supriya.ugens.Poll.ar(
            ...     label="Foo",
            ...     source=sine,
            ...     trigger=trigger,
            ...     trigger_id=1234,
            ... )
            >>> poll.label
            'Foo'

        Returns ugen input.
        """
        index = tuple(self._ordered_input_names).index("trigger_id") + 2
        characters = self._inputs[index:]
        characters = [chr(int(_)) for _ in characters]
        label = "".join(characters)
        return label


class RunningMax(Peak):
    """
    Tracks maximum signal amplitude.

    ::

        >>> source = supriya.ugens.In.ar(0)
        >>> trigger = supriya.ugens.Impulse.kr(1)
        >>> running_max = supriya.ugens.RunningMax.ar(
        ...     source=source,
        ...     trigger=0,
        ... )
        >>> running_max
        RunningMax.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None), ("trigger", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class RunningMin(Peak):
    """
    Tracks minimum signal amplitude.

    ::

        >>> source = supriya.ugens.In.ar(0)
        >>> trigger = supriya.ugens.Impulse.kr(1)
        >>> running_min = supriya.ugens.RunningMin.ar(
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> running_min
        RunningMin.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None), ("trigger", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Schmidt(UGen):
    """
    A Schmidt trigger.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> schmidt = supriya.ugens.Schmidt.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> schmidt
        Schmidt.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", 0), ("minimum", 0), ("maximum", 1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class SendPeakRMS(UGen):
    """
    Tracks peak and power of a signal for GUI applications.

    ::

        >>> source = supriya.ugens.In.ar(channel_count=4)
        >>> send_peak_rms = supriya.ugens.SendPeakRMS.kr(
        ...     command_name="/reply",
        ...     peak_lag=3,
        ...     reply_id=-1,
        ...     reply_rate=20,
        ...     source=source,
        ... )
        >>> send_peak_rms
        SendPeakRMS.kr()

    """

    ### CLASS VARIABLES ###

    _default_channel_count = 0
    _ordered_input_names = collections.OrderedDict(
        [("reply_rate", 20), ("peak_lag", 3), ("reply_id", -1)]
    )
    _unexpanded_argument_names = ("source",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### INITIALIZER ###

    def __init__(
        self,
        calculation_rate=None,
        command_name="/reply",
        peak_lag=3,
        reply_id=-1,
        reply_rate=20,
        source=None,
    ):
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            peak_lag=peak_lag,
            reply_id=reply_id,
            reply_rate=reply_rate,
        )
        command_name = str(command_name)
        if not isinstance(source, Sequence):
            source = (source,)
        self._configure_input("source", len(source))
        for input_ in source:
            self._configure_input("source", input_)
        self._configure_input("command_name", len(command_name))
        for character in command_name:
            self._configure_input("label", ord(character))

    ### PUBLIC METHODS ###

    @classmethod
    def ar(
        cls, command_name="/reply", peak_lag=3, reply_id=-1, reply_rate=20, source=None
    ):
        """
        Constructs an audio-rate SendPeakRMS.

        ::

            >>> source = supriya.ugens.In.ar(channel_count=4)
            >>> send_peak_rms = supriya.ugens.SendPeakRMS.ar(
            ...     command_name="/reply",
            ...     peak_lag=3,
            ...     reply_id=-1,
            ...     reply_rate=20,
            ...     source=source,
            ... )
            >>> send_peak_rms
            SendPeakRMS.ar()

        Returns ugen graph.
        """
        calculation_rate = CalculationRate.AUDIO
        ugen = cls._new_single(
            calculation_rate=calculation_rate,
            command_name=command_name,
            peak_lag=peak_lag,
            reply_id=reply_id,
            reply_rate=reply_rate,
            source=source,
        )
        return ugen

    @classmethod
    def kr(
        cls, command_name="/reply", peak_lag=3, reply_id=-1, reply_rate=20, source=None
    ):
        """
        Constructs a control-rate SendPeakRMS.

        ::

            >>> source = supriya.ugens.In.ar(channel_count=4)
            >>> send_peak_rms = supriya.ugens.SendPeakRMS.kr(
            ...     command_name="/reply",
            ...     peak_lag=3,
            ...     reply_id=-1,
            ...     reply_rate=20,
            ...     source=source,
            ... )
            >>> send_peak_rms
            SendPeakRMS.kr()

        Returns ugen graph.
        """
        calculation_rate = CalculationRate.CONTROL
        ugen = cls._new_single(
            calculation_rate=calculation_rate,
            command_name=command_name,
            peak_lag=peak_lag,
            reply_id=reply_id,
            reply_rate=reply_rate,
            source=source,
        )
        return ugen

    ### PUBLIC PROPERTIES ###

    @property
    def command_name(self):
        """
        Gets `command_name` input of SendPeakRMS.

        ::

            >>> source = supriya.ugens.In.ar(channel_count=4)
            >>> send_peak_rms = supriya.ugens.SendPeakRMS.ar(
            ...     command_name="/reply",
            ...     peak_lag=3,
            ...     reply_id=-1,
            ...     reply_rate=20,
            ...     source=source,
            ... )
            >>> send_peak_rms.command_name
            '/reply'

        Returns ugen input.
        """
        index = tuple(self._ordered_input_names).index("reply_id") + 1
        source_length = int(self._inputs[index])
        index += source_length + 2
        characters = self._inputs[index:]
        characters = [chr(int(_)) for _ in characters]
        command_name = "".join(characters)
        return command_name

    @property
    def source(self):
        """
        Gets `source` input of SendPeakRMS.

        ::

            >>> source = supriya.ugens.In.ar(channel_count=4)
            >>> send_peak_rms = supriya.ugens.SendPeakRMS.ar(
            ...     command_name="/reply",
            ...     peak_lag=3,
            ...     reply_id=-1,
            ...     reply_rate=20,
            ...     source=source,
            ... )
            >>> send_peak_rms.source
            (In.ar()[0], In.ar()[1], In.ar()[2], In.ar()[3])

        Returns ugen input.
        """
        index = tuple(self._ordered_input_names).index("reply_id") + 1
        source_length = int(self._inputs[index])
        start = index + 1
        stop = start + source_length
        return tuple(self._inputs[start:stop])


class SendTrig(UGen):
    _ordered_input_names = collections.OrderedDict(
        [("trigger", None), ("id_", 0), ("value", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Sweep(UGen):
    """
    A triggered linear ramp.

    ::

        >>> sweep = supriya.ugens.Sweep.ar(
        ...     rate=1,
        ...     trigger=0,
        ... )
        >>> sweep
        Sweep.ar()

    """

    _ordered_input_names = collections.OrderedDict([("trigger", 0), ("rate", 1)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class TDelay(UGen):
    """
    A trigger delay.

    ::

        >>> source = supriya.ugens.Dust.kr()
        >>> tdelay = supriya.ugens.TDelay.ar(
        ...     duration=0.1,
        ...     source=source,
        ... )
        >>> tdelay
        TDelay.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("duration", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class ToggleFF(UGen):
    """
    A toggle flip-flop.

    ::

        >>> trigger = supriya.ugens.Dust.kr(1)
        >>> toggle_ff = supriya.ugens.ToggleFF.ar(
        ...     trigger=trigger,
        ... )
        >>> toggle_ff
        ToggleFF.ar()

    """

    _ordered_input_names = collections.OrderedDict([("trigger", 0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Trig1(UGen):
    """
    A timed trigger.

    ::

        >>> source = supriya.ugens.Dust.kr(1)
        >>> trig_1 = supriya.ugens.Trig1.ar(
        ...     duration=0.1,
        ...     source=source,
        ... )
        >>> trig_1
        Trig1.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("duration", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Trig(UGen):
    """
    A timed trigger.

    ::

        >>> source = supriya.ugens.Dust.kr(1)
        >>> trig = supriya.ugens.Trig.ar(
        ...     duration=0.1,
        ...     source=source,
        ... )
        >>> trig
        Trig.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("duration", 0.1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Wrap(UGen):
    """
    Wraps a signal outside given thresholds.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> wrap = supriya.ugens.Wrap.ar(
        ...     maximum=0.9,
        ...     minimum=0.1,
        ...     source=source,
        ... )
        >>> wrap
        Wrap.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("source", 0), ("minimum", 0), ("maximum", 1)]
    )
    _valid_calculation_rates = (
        CalculationRate.AUDIO,
        CalculationRate.CONTROL,
        CalculationRate.SCALAR,
    )


class ZeroCrossing(UGen):
    """
    A zero-crossing frequency follower.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> zero_crossing = supriya.ugens.ZeroCrossing.ar(
        ...     source=source,
        ... )
        >>> zero_crossing
        ZeroCrossing.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)
